# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import operator

import numba
import numba.core.typing
import numba.core.typing.ctypes_utils

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()

########## for code that's built up from strings


def code_to_function(code, function_name, externals=None, debug=False):
    if debug:
        print("################### " + function_name)  # noqa: T001
        print(code)  # noqa: T001
    namespace = {} if externals is None else dict(externals)
    exec(code, namespace)
    return namespace[function_name]


########## Lookup


class Lookup(object):
    def __init__(self, layout):
        positions = []
        sharedptrs = []
        arrays = []
        tolookup(layout, positions, sharedptrs, arrays)
        assert len(positions) == len(sharedptrs)

        def find(x):
            for i, array in enumerate(arrays):
                if x is array:
                    return i
            else:
                assert isinstance(x, int)
                return x

        self.original_positions = positions
        self.positions = [find(x) for x in positions]
        self.sharedptrs_hold = sharedptrs
        self.arrays = arrays

        def arrayptr(x):
            if isinstance(x, int):
                return x
            else:
                return x.ctypes.data

        def sharedptr(x):
            if x is None:
                return -1
            elif x == 0:
                return 0
            else:
                return x.ptr()

        self.nplike = ak.nplike.of(layout)

        self.arrayptrs = self.nplike.array(
            [arrayptr(x) for x in positions], dtype=np.intp
        )
        self.sharedptrs = self.nplike.array(
            [sharedptr(x) for x in sharedptrs], dtype=np.intp
        )

    def _view_as_array(self):
        return self.nplike.vstack(
            [self.nplike.arange(len(self.arrayptrs)), self.arrayptrs, self.sharedptrs]
        ).T


def tolookup(layout, positions, sharedptrs, arrays):
    if isinstance(layout, ak.layout.NumpyArray):
        return ak._connect._numba.layout.NumpyArrayType.tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.forms.NumpyForm):
        return ak._connect._numba.layout.NumpyArrayType.form_tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.layout.RegularArray):
        return ak._connect._numba.layout.RegularArrayType.tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.forms.RegularForm):
        return ak._connect._numba.layout.RegularArrayType.form_tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(
        layout,
        (
            ak.layout.ListArray32,
            ak.layout.ListArrayU32,
            ak.layout.ListArray64,
            ak.layout.ListOffsetArray32,
            ak.layout.ListOffsetArrayU32,
            ak.layout.ListOffsetArray64,
        ),
    ):
        return ak._connect._numba.layout.ListArrayType.tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, (ak.forms.ListForm, ak.forms.ListOffsetForm)):
        return ak._connect._numba.layout.ListArrayType.form_tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(
        layout,
        (
            ak.layout.IndexedArray32,
            ak.layout.IndexedArrayU32,
            ak.layout.IndexedArray64,
        ),
    ):
        return ak._connect._numba.layout.IndexedArrayType.tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.forms.IndexedForm):
        return ak._connect._numba.layout.IndexedArrayType.form_tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(
        layout,
        (ak.layout.IndexedOptionArray32, ak.layout.IndexedOptionArray64),
    ):
        return ak._connect._numba.layout.IndexedOptionArrayType.tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.forms.IndexedOptionForm):
        return ak._connect._numba.layout.IndexedOptionArrayType.form_tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.layout.ByteMaskedArray):
        return ak._connect._numba.layout.ByteMaskedArrayType.tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.forms.ByteMaskedForm):
        return ak._connect._numba.layout.ByteMaskedArrayType.form_tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.layout.BitMaskedArray):
        return ak._connect._numba.layout.BitMaskedArrayType.tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.forms.BitMaskedForm):
        return ak._connect._numba.layout.BitMaskedArrayType.form_tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.layout.UnmaskedArray):
        return ak._connect._numba.layout.UnmaskedArrayType.tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.forms.UnmaskedForm):
        return ak._connect._numba.layout.UnmaskedArrayType.form_tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.layout.RecordArray):
        return ak._connect._numba.layout.RecordArrayType.tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.forms.RecordForm):
        return ak._connect._numba.layout.RecordArrayType.form_tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.layout.Record):
        return ak._connect._numba.layout.RecordType.tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(
        layout,
        (
            ak.layout.UnionArray8_32,
            ak.layout.UnionArray8_U32,
            ak.layout.UnionArray8_64,
        ),
    ):
        return ak._connect._numba.layout.UnionArrayType.tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.forms.UnionForm):
        return ak._connect._numba.layout.UnionArrayType.form_tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.layout.VirtualArray):
        return ak._connect._numba.layout.VirtualArrayType.tolookup(
            layout, positions, sharedptrs, arrays
        )

    elif isinstance(layout, ak.forms.VirtualForm):
        return ak._connect._numba.layout.VirtualArrayType.form_tolookup(
            layout, positions, sharedptrs, arrays
        )

    else:
        raise AssertionError(
            "unrecognized Content or Form type: {0}".format(type(layout))
            + ak._util.exception_suffix(__file__)
        )


def tonumbatype(form):
    if isinstance(form, ak.forms.NumpyForm):
        return ak._connect._numba.layout.NumpyArrayType.from_form(form)

    elif isinstance(form, ak.forms.RegularForm):
        return ak._connect._numba.layout.RegularArrayType.from_form(form)

    elif isinstance(form, (ak.forms.ListForm, ak.forms.ListOffsetForm)):
        return ak._connect._numba.layout.ListArrayType.from_form(form)

    elif isinstance(form, ak.forms.IndexedForm):
        return ak._connect._numba.layout.IndexedArrayType.from_form(form)

    elif isinstance(form, ak.forms.IndexedOptionForm):
        return ak._connect._numba.layout.IndexedOptionArrayType.from_form(form)

    elif isinstance(form, ak.forms.ByteMaskedForm):
        return ak._connect._numba.layout.ByteMaskedArrayType.from_form(form)

    elif isinstance(form, ak.forms.BitMaskedForm):
        return ak._connect._numba.layout.BitMaskedArrayType.from_form(form)

    elif isinstance(form, ak.forms.UnmaskedForm):
        return ak._connect._numba.layout.UnmaskedArrayType.from_form(form)

    elif isinstance(form, ak.forms.RecordForm):
        return ak._connect._numba.layout.RecordArrayType.from_form(form)

    elif isinstance(form, ak.forms.UnionForm):
        return ak._connect._numba.layout.UnionArrayType.from_form(form)

    elif isinstance(form, ak.forms.VirtualForm):
        return ak._connect._numba.layout.VirtualArrayType.from_form(form)

    else:
        raise AssertionError(
            "unrecognized Form type: {0}".format(type(form))
            + ak._util.exception_suffix(__file__)
        )


@numba.extending.typeof_impl.register(Lookup)
def typeof_Lookup(obj, c):
    return LookupType()


class LookupType(numba.types.Type):
    arraytype = numba.types.Array(numba.intp, 1, "C")

    def __init__(self):
        super(LookupType, self).__init__(name="ak.LookupType()")


@numba.extending.register_model(LookupType)
class LookupModel(numba.core.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("arrayptrs", fe_type.arraytype), ("sharedptrs", fe_type.arraytype)]
        super(LookupModel, self).__init__(dmm, fe_type, members)


@numba.extending.unbox(LookupType)
def unbox_Lookup(lookuptype, lookupobj, c):
    arrayptrs_obj = c.pyapi.object_getattr_string(lookupobj, "arrayptrs")
    sharedptrs_obj = c.pyapi.object_getattr_string(lookupobj, "sharedptrs")

    proxyout = c.context.make_helper(c.builder, lookuptype)
    proxyout.arrayptrs = c.pyapi.to_native_value(
        lookuptype.arraytype, arrayptrs_obj
    ).value
    proxyout.sharedptrs = c.pyapi.to_native_value(
        lookuptype.arraytype, sharedptrs_obj
    ).value

    c.pyapi.decref(arrayptrs_obj)
    c.pyapi.decref(sharedptrs_obj)

    is_error = numba.core.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)


########## ArrayView


class ArrayView(object):
    @classmethod
    def fromarray(cls, array):
        behavior = ak._util.behaviorof(array)
        layout = ak.operations.convert.to_layout(
            array,
            allow_record=False,
            allow_other=False,
            numpytype=(np.number, bool, np.bool_),
        )
        while isinstance(layout, ak.layout.VirtualArray) and isinstance(
            layout.generator, ak.layout.SliceGenerator
        ):
            layout = layout.array
        layout = ak.operations.convert.regularize_numpyarray(
            layout, allow_empty=False, highlevel=False
        )

        if isinstance(layout, ak.partition.PartitionedArray):
            numba_type = None
            for part in layout.partitions:
                if numba_type is None:
                    numba_type = ak._connect._numba.layout.typeof(part)
                elif numba_type != ak._connect._numba.layout.typeof(part):
                    raise ValueError(
                        "partitioned arrays can only be used in Numba if all "
                        "partitions have the same numba_type"
                        + ak._util.exception_suffix(__file__)
                    )
            return PartitionedView(
                ak._connect._numba.layout.typeof(part),
                behavior,
                [Lookup(x) for x in layout.partitions],
                ak.nplike.of(layout).asarray(layout.stops, dtype=np.intp),
                0,
                len(layout),
                (),
            )

        else:
            return ArrayView(
                ak._connect._numba.layout.typeof(layout),
                behavior,
                Lookup(layout),
                0,
                0,
                len(layout),
                (),
            )

    def __init__(self, type, behavior, lookup, pos, start, stop, fields):
        self.type = type
        self.behavior = behavior
        self.lookup = lookup
        self.pos = pos
        self.start = start
        self.stop = stop
        self.fields = fields

    def toarray(self):
        layout = self.type.tolayout(self.lookup, self.pos, self.fields)
        sliced = layout.getitem_range_nowrap(self.start, self.stop)
        return ak._util.wrap(sliced, self.behavior)


@numba.extending.typeof_impl.register(ArrayView)
def typeof_ArrayView(obj, c):
    return ArrayViewType(obj.type, obj.behavior, obj.fields)


def wrap(type, viewtype, fields):
    if fields is None:
        return ArrayViewType(type, viewtype.behavior, viewtype.fields)
    else:
        return ArrayViewType(type, viewtype.behavior, fields)


class ArrayViewType(numba.types.IterableType, numba.types.Sized):
    def __init__(self, type, behavior, fields):
        super(ArrayViewType, self).__init__(
            name="ak.ArrayView({0}, {1}, {2})".format(
                type.name,
                ak._connect._numba.repr_behavior(behavior),
                repr(fields),
            )
        )
        self.type = type
        self.behavior = behavior
        self.fields = fields

    @property
    def iterator_type(self):
        return IteratorType(self)


@numba.extending.register_model(ArrayViewType)
class ArrayViewModel(numba.core.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ("pos", numba.intp),
            ("start", numba.intp),
            ("stop", numba.intp),
            ("arrayptrs", numba.types.CPointer(numba.intp)),
            ("sharedptrs", numba.types.CPointer(numba.intp)),
            ("pylookup", numba.types.pyobject),
        ]
        super(ArrayViewModel, self).__init__(dmm, fe_type, members)


@numba.core.imputils.lower_constant(ArrayViewType)
def lower_const_Array(context, builder, viewtype, array):
    return lower_const_view(context, builder, viewtype, array._numbaview)


def lower_const_view(context, builder, viewtype, view):
    lookup = view.lookup
    arrayptrs = lookup.arrayptrs
    sharedptrs = lookup.sharedptrs
    pos = view.pos
    start = view.start
    stop = view.stop

    arrayptrs_val = context.make_constant_array(
        builder, numba.typeof(arrayptrs), arrayptrs
    )
    sharedptrs_val = context.make_constant_array(
        builder, numba.typeof(sharedptrs), sharedptrs
    )

    proxyout = context.make_helper(builder, viewtype)
    proxyout.pos = context.get_constant(numba.intp, pos)
    proxyout.start = context.get_constant(numba.intp, start)
    proxyout.stop = context.get_constant(numba.intp, stop)
    proxyout.arrayptrs = context.make_helper(
        builder, numba.typeof(arrayptrs), arrayptrs_val
    ).data
    proxyout.sharedptrs = context.make_helper(
        builder, numba.typeof(sharedptrs), sharedptrs_val
    ).data
    proxyout.pylookup = context.add_dynamic_addr(
        builder, id(lookup), info=str(type(lookup))
    )

    return proxyout._getvalue()


@numba.extending.unbox(ArrayViewType)
def unbox_Array(viewtype, arrayobj, c):
    view_obj = c.pyapi.object_getattr_string(arrayobj, "_numbaview")
    out = unbox_ArrayView(viewtype, view_obj, c)
    c.pyapi.decref(view_obj)
    return out


def unbox_ArrayView(viewtype, view_obj, c):
    lookup_obj = c.pyapi.object_getattr_string(view_obj, "lookup")
    pos_obj = c.pyapi.object_getattr_string(view_obj, "pos")
    start_obj = c.pyapi.object_getattr_string(view_obj, "start")
    stop_obj = c.pyapi.object_getattr_string(view_obj, "stop")

    lookup_val = c.pyapi.to_native_value(LookupType(), lookup_obj).value
    lookup_proxy = c.context.make_helper(c.builder, LookupType(), lookup_val)

    proxyout = c.context.make_helper(c.builder, viewtype)
    proxyout.pos = c.pyapi.number_as_ssize_t(pos_obj)
    proxyout.start = c.pyapi.number_as_ssize_t(start_obj)
    proxyout.stop = c.pyapi.number_as_ssize_t(stop_obj)
    proxyout.arrayptrs = c.context.make_helper(
        c.builder, LookupType.arraytype, lookup_proxy.arrayptrs
    ).data
    proxyout.sharedptrs = c.context.make_helper(
        c.builder, LookupType.arraytype, lookup_proxy.sharedptrs
    ).data
    proxyout.pylookup = lookup_obj

    c.pyapi.decref(lookup_obj)
    c.pyapi.decref(pos_obj)
    c.pyapi.decref(start_obj)
    c.pyapi.decref(stop_obj)

    if c.context.enable_nrt:
        c.context.nrt.decref(c.builder, LookupType(), lookup_val)

    is_error = numba.core.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)


@numba.extending.box(ArrayViewType)
def box_Array(viewtype, viewval, c):
    arrayview_obj = box_ArrayView(viewtype, viewval, c)
    out = c.pyapi.call_method(arrayview_obj, "toarray", ())
    c.pyapi.decref(arrayview_obj)
    return out


def dict2serializable(obj):
    if obj is None:
        return None
    else:
        return tuple(obj.items())


def serializable2dict(obj):
    if obj is None:
        return None
    else:
        return dict(obj)


def box_ArrayView(viewtype, viewval, c):
    serializable2dict_obj = c.pyapi.unserialize(
        c.pyapi.serialize_object(serializable2dict)
    )
    behavior2_obj = c.pyapi.unserialize(
        c.pyapi.serialize_object(dict2serializable(viewtype.behavior))
    )
    behavior_obj = c.pyapi.call_function_objargs(
        serializable2dict_obj, (behavior2_obj,)
    )
    ArrayView_obj = c.pyapi.unserialize(c.pyapi.serialize_object(ArrayView))
    type_obj = c.pyapi.unserialize(c.pyapi.serialize_object(viewtype.type))
    fields_obj = c.pyapi.unserialize(c.pyapi.serialize_object(viewtype.fields))

    proxyin = c.context.make_helper(c.builder, viewtype, viewval)
    pos_obj = c.pyapi.long_from_ssize_t(proxyin.pos)
    start_obj = c.pyapi.long_from_ssize_t(proxyin.start)
    stop_obj = c.pyapi.long_from_ssize_t(proxyin.stop)
    lookup_obj = proxyin.pylookup

    out = c.pyapi.call_function_objargs(
        ArrayView_obj,
        (type_obj, behavior_obj, lookup_obj, pos_obj, start_obj, stop_obj, fields_obj),
    )

    c.pyapi.decref(serializable2dict_obj)
    c.pyapi.decref(behavior2_obj)
    c.pyapi.decref(behavior_obj)
    c.pyapi.decref(ArrayView_obj)
    c.pyapi.decref(type_obj)
    c.pyapi.decref(fields_obj)
    c.pyapi.decref(pos_obj)
    c.pyapi.decref(start_obj)
    c.pyapi.decref(stop_obj)

    return out


@numba.core.typing.templates.infer_global(len)
class type_len(numba.core.typing.templates.AbstractTemplate):
    def generic(self, args, kwargs):
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], ArrayViewType):
            return numba.intp(args[0])


@numba.extending.lower_builtin(len, ArrayViewType)
def lower_len(context, builder, sig, args):
    proxyin = context.make_helper(builder, sig.args[0], args[0])
    return builder.sub(proxyin.stop, proxyin.start)


@numba.core.typing.templates.infer_global(operator.getitem)
class type_getitem(numba.core.typing.templates.AbstractTemplate):
    def generic(self, args, kwargs):
        if len(args) == 2 and len(kwargs) == 0 and isinstance(args[0], ArrayViewType):
            viewtype, wheretype = args
            if isinstance(wheretype, numba.types.Integer):
                return viewtype.type.getitem_at_check(viewtype)(viewtype, wheretype)
            elif (
                isinstance(wheretype, numba.types.SliceType) and not wheretype.has_step
            ):
                return viewtype.type.getitem_range(viewtype)(viewtype, wheretype)
            elif isinstance(wheretype, numba.types.StringLiteral):
                return viewtype.type.getitem_field(viewtype, wheretype.literal_value)(
                    viewtype, wheretype
                )
            else:
                raise TypeError(
                    "only an integer, start:stop range, or a *constant* "
                    "field name string may be used as ak.Array "
                    "slices in compiled code" + ak._util.exception_suffix(__file__)
                )


@numba.extending.lower_builtin(operator.getitem, ArrayViewType, numba.types.Integer)
def lower_getitem_at(context, builder, sig, args):
    rettype, (viewtype, wheretype) = sig.return_type, sig.args
    viewval, whereval = args
    viewproxy = context.make_helper(builder, viewtype, viewval)
    return viewtype.type.lower_getitem_at_check(
        context,
        builder,
        rettype,
        viewtype,
        viewval,
        viewproxy,
        wheretype,
        whereval,
        True,
        True,
    )


@numba.extending.lower_builtin(operator.getitem, ArrayViewType, numba.types.slice2_type)
def lower_getitem_range(context, builder, sig, args):
    rettype, (viewtype, wheretype) = sig.return_type, sig.args
    viewval, whereval = args
    viewproxy = context.make_helper(builder, viewtype, viewval)
    whereproxy = context.make_helper(builder, wheretype, whereval)
    return viewtype.type.lower_getitem_range(
        context,
        builder,
        rettype,
        viewtype,
        viewval,
        viewproxy,
        whereproxy.start,
        whereproxy.stop,
        True,
    )


@numba.extending.lower_builtin(
    operator.getitem, ArrayViewType, numba.types.StringLiteral
)
def lower_getitem_field(context, builder, sig, args):
    _, (viewtype, wheretype) = sig.return_type, sig.args
    viewval, whereval = args
    return viewtype.type.lower_getitem_field(
        context, builder, viewtype, viewval, wheretype.literal_value
    )


@numba.core.typing.templates.infer_getattr
class type_getattr(numba.core.typing.templates.AttributeTemplate):
    key = ArrayViewType

    def generic_resolve(self, viewtype, attr):
        if attr == "ndim":
            return numba.intp
        else:
            return viewtype.type.getitem_field(viewtype, attr)


@numba.extending.lower_getattr_generic(ArrayViewType)
def lower_getattr_generic(context, builder, viewtype, viewval, attr):
    if attr == "ndim":
        return context.get_constant(numba.intp, viewtype.type.ndim)
    else:
        return viewtype.type.lower_getitem_field(
            context, builder, viewtype, viewval, attr
        )


class IteratorType(numba.types.common.SimpleIteratorType):
    def __init__(self, viewtype):
        super(IteratorType, self).__init__(
            "ak.Iterator({0})".format(viewtype.name),
            viewtype.type.getitem_at_check(viewtype),
        )
        self.viewtype = viewtype


@numba.core.typing.templates.infer
class type_getiter(numba.core.typing.templates.AbstractTemplate):
    key = "getiter"

    def generic(self, args, kwargs):
        if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], ArrayViewType):
            return IteratorType(args[0])(args[0])


@numba.core.datamodel.registry.register_default(IteratorType)
class IteratorModel(numba.core.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ("view", fe_type.viewtype),
            ("length", numba.intp),
            ("at", numba.types.EphemeralPointer(numba.intp)),
        ]
        super(IteratorModel, self).__init__(dmm, fe_type, members)


@numba.extending.lower_builtin("getiter", ArrayViewType)
def lower_getiter(context, builder, sig, args):
    rettype, (viewtype,) = sig.return_type, sig.args
    (viewval,) = args
    viewproxy = context.make_helper(builder, viewtype, viewval)
    proxyout = context.make_helper(builder, rettype)
    proxyout.view = viewval
    proxyout.length = builder.sub(viewproxy.stop, viewproxy.start)
    proxyout.at = numba.core.cgutils.alloca_once_value(
        builder, context.get_constant(numba.intp, 0)
    )
    if context.enable_nrt:
        context.nrt.incref(builder, viewtype, viewval)
    return numba.core.imputils.impl_ret_new_ref(
        context, builder, rettype, proxyout._getvalue()
    )


@numba.extending.lower_builtin("iternext", IteratorType)
@numba.core.imputils.iternext_impl(numba.core.imputils.RefType.BORROWED)
def lower_iternext(context, builder, sig, args, result):
    (itertype,) = sig.args
    (iterval,) = args
    proxyin = context.make_helper(builder, itertype, iterval)
    at = builder.load(proxyin.at)

    is_valid = builder.icmp_signed("<", at, proxyin.length)
    result.set_valid(is_valid)

    with builder.if_then(is_valid, likely=True):
        result.yield_(
            lower_getitem_at(
                context,
                builder,
                itertype.yield_type(itertype.viewtype, numba.intp),
                (proxyin.view, at),
            )
        )
        nextat = numba.core.cgutils.increment_index(builder, at)
        builder.store(nextat, proxyin.at)


########## RecordView


class RecordView(object):
    @classmethod
    def fromrecord(cls, record):
        behavior = ak._util.behaviorof(record)
        layout = ak.operations.convert.to_layout(
            record,
            allow_record=True,
            allow_other=False,
            numpytype=(np.number, bool, np.bool_),
        )
        assert isinstance(layout, ak.layout.Record)
        arraylayout = layout.array
        return RecordView(
            ArrayView(
                ak._connect._numba.layout.typeof(arraylayout),
                behavior,
                Lookup(arraylayout),
                0,
                0,
                len(arraylayout),
                (),
            ),
            layout.at,
        )

    def __init__(self, arrayview, at):
        self.arrayview = arrayview
        self.at = at

    def torecord(self):
        arraylayout = self.arrayview.toarray().layout
        return ak._util.wrap(
            ak.layout.Record(arraylayout, self.at), self.arrayview.behavior
        )


@numba.extending.typeof_impl.register(RecordView)
def typeof_RecordView(obj, c):
    return RecordViewType(numba.typeof(obj.arrayview))


class RecordViewType(numba.types.Type):
    def __init__(self, arrayviewtype):
        super(RecordViewType, self).__init__(
            name="ak.RecordViewType({0})".format(arrayviewtype.name)
        )
        self.arrayviewtype = arrayviewtype

    @property
    def behavior(self):
        return self.arrayviewtype.behavior

    @property
    def fields(self):
        return self.arrayviewtype.fields

    def typer_field(self, key):
        return self.arrayviewtype.type.getitem_field_record(self, key)

    def lower_field(self, context, builder, val, key):
        return self.arrayviewtype.type.lower_getitem_field_record(
            context, builder, self, val, key
        )


@numba.extending.register_model(RecordViewType)
class RecordViewModel(numba.core.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [("arrayview", fe_type.arrayviewtype), ("at", numba.intp)]
        super(RecordViewModel, self).__init__(dmm, fe_type, members)


@numba.core.imputils.lower_constant(RecordViewType)
def lower_const_Record(context, builder, recordviewtype, record):
    arrayview_val = lower_const_view(
        context, builder, recordviewtype.arrayviewtype, record._numbaview.arrayview
    )
    proxyout = context.make_helper(builder, recordviewtype)
    proxyout.arrayview = arrayview_val
    proxyout.at = context.get_constant(numba.intp, record._layout.at)
    return proxyout._getvalue()


@numba.extending.unbox(RecordViewType)
def unbox_RecordView(recordviewtype, recordobj, c):
    recordview_obj = c.pyapi.object_getattr_string(recordobj, "_numbaview")
    arrayview_obj = c.pyapi.object_getattr_string(recordview_obj, "arrayview")
    at_obj = c.pyapi.object_getattr_string(recordview_obj, "at")

    arrayview_val = unbox_ArrayView(
        recordviewtype.arrayviewtype, arrayview_obj, c
    ).value

    proxyout = c.context.make_helper(c.builder, recordviewtype)
    proxyout.arrayview = arrayview_val
    proxyout.at = c.pyapi.number_as_ssize_t(at_obj)

    c.pyapi.decref(recordview_obj)
    c.pyapi.decref(at_obj)

    if c.context.enable_nrt:
        c.context.nrt.decref(c.builder, recordviewtype.arrayviewtype, arrayview_val)

    is_error = numba.core.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)


@numba.extending.box(RecordViewType)
def box_RecordView(recordviewtype, viewval, c):
    RecordView_obj = c.pyapi.unserialize(c.pyapi.serialize_object(RecordView))

    proxyin = c.context.make_helper(c.builder, recordviewtype, viewval)
    arrayview_obj = box_ArrayView(recordviewtype.arrayviewtype, proxyin.arrayview, c)
    at_obj = c.pyapi.long_from_ssize_t(proxyin.at)

    recordview_obj = c.pyapi.call_function_objargs(
        RecordView_obj, (arrayview_obj, at_obj)
    )

    out = c.pyapi.call_method(recordview_obj, "torecord", ())

    c.pyapi.decref(RecordView_obj)
    c.pyapi.decref(arrayview_obj)
    c.pyapi.decref(at_obj)
    c.pyapi.decref(recordview_obj)

    return out


@numba.core.typing.templates.infer_global(operator.getitem)
class type_getitem_record(numba.core.typing.templates.AbstractTemplate):
    def generic(self, args, kwargs):
        if len(args) == 2 and len(kwargs) == 0 and isinstance(args[0], RecordViewType):
            recordviewtype, wheretype = args
            if isinstance(wheretype, numba.types.StringLiteral):
                return recordviewtype.arrayviewtype.type.getitem_field_record(
                    recordviewtype, wheretype.literal_value
                )(recordviewtype, wheretype)
            else:
                raise TypeError(
                    "only a *constant* field name string may be used as "
                    "ak.Record slices in compiled code"
                    + ak._util.exception_suffix(__file__)
                )


@numba.extending.lower_builtin(
    operator.getitem, RecordViewType, numba.types.StringLiteral
)
def lower_getitem_field_record(context, builder, sig, args):
    _, (recordviewtype, wheretype) = sig.return_type, sig.args
    recordviewval, whereval = args
    return recordviewtype.arrayviewtype.type.lower_getitem_field_record(
        context, builder, recordviewtype, recordviewval, wheretype.literal_value
    )


@numba.core.typing.templates.infer_getattr
class type_getattr_record(numba.core.typing.templates.AttributeTemplate):
    key = RecordViewType

    def generic_resolve(self, recordviewtype, attr):
        for methodname, typer, lower in ak._util.numba_methods(
            recordviewtype.arrayviewtype.type, recordviewtype.arrayviewtype.behavior
        ):
            if attr == methodname:

                class type_method(numba.core.typing.templates.AbstractTemplate):
                    key = methodname

                    def generic(self, args, kwargs):
                        if len(kwargs) == 0:
                            sig = typer(recordviewtype, args)
                            sig = numba.core.typing.templates.Signature(
                                sig.return_type, sig.args, recordviewtype
                            )
                            numba.extending.lower_builtin(
                                methodname,
                                recordviewtype,
                                *[
                                    x.literal_type
                                    if isinstance(x, numba.types.Literal)
                                    else x
                                    for x in args
                                ]
                            )(lower)
                            return sig

                return numba.types.BoundFunction(type_method, recordviewtype)

        for attrname, typer, _ in ak._util.numba_attrs(
            recordviewtype.arrayviewtype.type, recordviewtype.arrayviewtype.behavior
        ):
            if attr == attrname:
                return typer(recordviewtype)

        else:
            return recordviewtype.typer_field(attr)


@numba.extending.lower_getattr_generic(RecordViewType)
def lower_getattr_generic_record(context, builder, recordviewtype, recordviewval, attr):
    for attrname, typer, lower in ak._util.numba_attrs(
        recordviewtype.arrayviewtype.type, recordviewtype.arrayviewtype.behavior
    ):
        if attr == attrname:
            return lower(
                context,
                builder,
                typer(recordviewtype)(recordviewtype),
                (recordviewval,),
            )
    else:
        return recordviewtype.lower_field(context, builder, recordviewval, attr)


def register_unary_operator(unaryop):
    @numba.core.typing.templates.infer_global(unaryop)
    class type_binary_operator(numba.core.typing.templates.AbstractTemplate):
        def generic(self, args, kwargs):
            if len(args) == 1 and len(kwargs) == 0:
                behavior = None

                if isinstance(args[0], RecordViewType):
                    left = args[0].arrayviewtype.type
                    behavior = args[0].arrayviewtype.behavior

                    for typer, lower in ak._util.numba_unaryops(
                        unaryop, left, behavior
                    ):
                        numba.extending.lower_builtin(unaryop, *args)(lower)
                        return typer(unaryop, args[0])


for unaryop in (
    abs,
    operator.inv,
    operator.invert,
    operator.neg,
    operator.not_,
    operator.pos,
    operator.truth,
):
    register_unary_operator(unaryop)


def register_binary_operator(binop):
    @numba.core.typing.templates.infer_global(binop)
    class type_binary_operator(numba.core.typing.templates.AbstractTemplate):
        def generic(self, args, kwargs):
            if len(args) == 2 and len(kwargs) == 0:
                left, right, behavior = None, None, None

                if isinstance(args[0], RecordViewType):
                    left = args[0].arrayviewtype.type
                    behavior = args[0].arrayviewtype.behavior

                if isinstance(args[1], RecordViewType):
                    right = args[1].arrayviewtype.type
                    if behavior is None:
                        behavior = args[1].arrayviewtype.behavior

                if left is not None or right is not None:
                    for typer, lower in ak._util.numba_binops(
                        binop, left, right, behavior
                    ):
                        numba.extending.lower_builtin(binop, *args)(lower)
                        return typer(binop, args[0], args[1])


for binop in (
    operator.add,
    operator.and_,
    operator.contains,
    operator.eq,
    operator.floordiv,
    operator.ge,
    operator.gt,
    operator.le,
    operator.lshift,
    operator.lt,
    operator.mod,
    operator.mul,
    operator.ne,
    operator.or_,
    operator.pow,
    operator.rshift,
    operator.sub,
    operator.truediv,
    operator.xor,
) + (() if not hasattr(operator, "matmul") else (operator.matmul,)):
    register_binary_operator(binop)


########## __contains__


@numba.extending.overload(operator.contains)
def overload_contains(obj, element):
    if isinstance(obj, (ArrayViewType, RecordViewType)) and (
        (element == numba.types.none)
        or (isinstance(element, (numba.types.Number, numba.types.Boolean)))
        or (
            isinstance(element, numba.types.Optional)
            and isinstance(element.type, (numba.types.Number, numba.types.Boolean))
        )
    ):
        statements = []

        def add_statement(indent, name, arraytype, is_array):
            if is_array:
                statements.append("for x in " + name + ":")
                name = "x"
                indent = indent + "    "

            if isinstance(arraytype, ak._connect._numba.layout.RecordArrayType):
                if arraytype.is_tuple:
                    for fi, ft in enumerate(arraytype.contenttypes):
                        add_statement(indent, name + "[" + repr(fi) + "]", ft, False)
                else:
                    for fn, ft in zip(arraytype.recordlookup, arraytype.contenttypes):
                        add_statement(indent, name + "[" + repr(fn) + "]", ft, False)

            elif arraytype.ndim == 1 and not arraytype.is_recordtype:
                if arraytype.is_optiontype:
                    statements.append(
                        indent + "if (element is None and {0} is None) or "
                        "({0} is not None and element == {0}): return True".format(name)
                    )
                else:
                    statements.append(
                        indent + "if element == {0}: return True".format(name)
                    )

            else:
                if arraytype.is_optiontype:
                    statements.append(
                        indent + "if (element is None and {0} is None) or "
                        "({0} is not None and element in {0}): return True".format(name)
                    )
                else:
                    statements.append(
                        indent + "if element in {0}: return True".format(name)
                    )

        if isinstance(obj, ArrayViewType):
            add_statement("", "obj", obj.type, True)
        else:
            add_statement("", "obj", obj.arrayviewtype.type, False)

        return code_to_function(
            """
def contains_impl(obj, element):
    {0}
    return False""".format(
                "\n    ".join(statements)
            ),
            "contains_impl",
        )


########## np.array and np.asarray


array_supported = (
    numba.types.boolean,
    numba.types.int8,
    numba.types.int16,
    numba.types.int32,
    numba.types.int64,
    numba.types.uint8,
    numba.types.uint16,
    numba.types.uint32,
    numba.types.uint64,
    numba.types.float32,
    numba.types.float64,
)


@numba.extending.overload(ak.nplike.numpy.array)
def overload_np_array(array, dtype=None):
    if isinstance(array, ArrayViewType):
        ndim = array.type.ndim
        inner_dtype = array.type.inner_dtype
        if ndim is not None and inner_dtype in array_supported:
            declare_shape = []
            compute_shape = []
            specify_shape = ["len(array)"]
            ensure_shape = []
            array_name = "array"
            for i in range(ndim - 1):
                declare_shape.append("shape{0} = -1".format(i))
                compute_shape.append(
                    "{0}for x{1} in {2}:".format("    " * i, i, array_name)
                )
                compute_shape.append("{0}    if shape{1} == -1:".format("    " * i, i))
                compute_shape.append(
                    "{0}        shape{1} = len(x{1})".format("    " * i, i)
                )
                compute_shape.append(
                    "{0}    elif shape{1} != len(x{1}):".format("    " * i, i)
                )
                compute_shape.append(
                    "{0}        raise ValueError('cannot convert to NumPy because "
                    "subarray lengths are not regular')".format("    " * i)
                )
                specify_shape.append("shape{0}".format(i))
                ensure_shape.append("if shape{0} == -1: shape{0} = 0".format(i))
                array_name = "x{0}".format(i)

            fill_array = []
            index = []
            array_name = "array"
            for i in range(ndim):
                fill_array.append(
                    "{0}for i{1}, x{1} in enumerate({2}):".format(
                        "    " * i, i, array_name
                    )
                )
                index.append("i{0}".format(i))
                array_name = "x{0}".format(i)

            fill_array.append(
                "{0}out[{1}] = x{2}".format("    " * ndim, "][".join(index), ndim - 1)
            )

            return code_to_function(
                """
def array_impl(array, dtype=None):
    {0}
    {1}
    {2}
    out = numpy.zeros(({3}), {4})
    {5}
    return out
""".format(
                    "\n    ".join(declare_shape),
                    "\n    ".join(compute_shape),
                    "\n    ".join(ensure_shape),
                    ", ".join(specify_shape),
                    "numpy.{0}".format(inner_dtype) if dtype is None else "dtype",
                    "\n    ".join(fill_array),
                ),
                "array_impl",
                {"numpy": ak.nplike.numpy},
            )


@numba.extending.type_callable(ak.nplike.numpy.asarray)
def type_asarray(context):
    def typer(arrayview):
        if (
            isinstance(arrayview, ArrayViewType)
            and isinstance(arrayview.type, ak._connect._numba.layout.NumpyArrayType)
            and arrayview.type.ndim == 1
            and arrayview.type.inner_dtype in array_supported
        ):
            return numba.types.Array(arrayview.type.inner_dtype, 1, "C")

    return typer


@numba.extending.lower_builtin(ak.nplike.numpy.asarray, ArrayViewType)
def lower_asarray(context, builder, sig, args):
    rettype, (viewtype,) = sig.return_type, sig.args
    (viewval,) = args
    viewproxy = context.make_helper(builder, viewtype, viewval)
    assert isinstance(viewtype.type, ak._connect._numba.layout.NumpyArrayType)

    whichpos = ak._connect._numba.layout.posat(
        context, builder, viewproxy.pos, viewtype.type.ARRAY
    )
    arrayptr = ak._connect._numba.layout.getat(
        context, builder, viewproxy.arrayptrs, whichpos
    )

    bitwidth = ak._connect._numba.layout.type_bitwidth(rettype.dtype)
    itemsize = context.get_constant(numba.intp, bitwidth // 8)

    data = numba.core.cgutils.pointer_add(
        builder,
        arrayptr,
        builder.mul(viewproxy.start, itemsize),
        context.get_value_type(numba.types.CPointer(rettype.dtype)),
    )

    shape = context.make_tuple(
        builder,
        numba.types.UniTuple(numba.types.intp, 1),
        (builder.sub(viewproxy.stop, viewproxy.start),),
    )
    strides = context.make_tuple(
        builder,
        numba.types.UniTuple(numba.types.intp, 1),
        (itemsize,),
    )

    out = numba.np.arrayobj.make_array(rettype)(context, builder)
    numba.np.arrayobj.populate_array(
        out,
        data=data,
        shape=shape,
        strides=strides,
        itemsize=itemsize,
        meminfo=None,
        parent=None,
    )
    return out._getvalue()


########## PartitionedView


class PartitionedView(object):
    def __init__(self, type, behavior, lookups, stops, start, stop, fields):
        self.type = type
        self.behavior = behavior
        self.lookups = lookups
        self.stops = stops
        self.start = start
        self.stop = stop
        self.fields = fields

    def toarray(self):
        output = []
        partition_start = 0
        for partitionid, lookup in enumerate(self.lookups):
            partition_stop = self.stops[partitionid]

            if partition_start <= self.start and self.stop <= partition_stop:
                layout = self.type.tolayout(lookup, 0, self.fields)
                output.append(
                    layout[self.start - partition_start : self.stop - partition_start]
                )
                break

            elif partition_start <= self.start < partition_stop:
                layout = self.type.tolayout(lookup, 0, self.fields)
                output.append(
                    layout[
                        self.start - partition_start : partition_stop - partition_start
                    ]
                )

            elif partition_start < self.stop <= partition_stop:
                layout = self.type.tolayout(lookup, 0, self.fields)
                output.append(layout[0 : self.stop - partition_start])
                break

            elif self.start < partition_start and partition_stop < self.stop:
                layout = self.type.tolayout(lookup, 0, self.fields)
                output.append(layout[0 : partition_stop - partition_start])

            partition_start = partition_stop

        return ak._util.wrap(
            ak.partition.IrregularlyPartitionedArray(output), self.behavior
        )


@numba.extending.typeof_impl.register(PartitionedView)
def typeof_PartitionedView(obj, c):
    return PartitionedViewType(obj.type, obj.behavior, obj.fields)


class PartitionedViewType(numba.types.IterableType, numba.types.Sized):
    stopstype = numba.types.Array(numba.intp, 1, "C")

    def __init__(self, type, behavior, fields):
        super(PartitionedViewType, self).__init__(
            name="ak.PartitionedView({0}, {1}, {2})".format(
                type.name,
                ak._connect._numba.repr_behavior(behavior),
                repr(fields),
            )
        )
        self.type = type
        self.behavior = behavior
        self.fields = fields

    @property
    def iterator_type(self):
        return PartitionedIteratorType(self)

    def toArrayViewType(self):
        return ArrayViewType(self.type, self.behavior, self.fields)

    def getitem_field(self, key):
        return PartitionedViewType(self.type, self.behavior, self.fields + (key,))

    def lower_get_localstart(self, context, builder, stops, partitionid):
        out = numba.core.cgutils.alloca_once_value(
            builder, context.get_constant(numba.intp, 0)
        )

        with builder.if_then(
            builder.icmp_signed("!=", partitionid, context.get_constant(numba.intp, 0))
        ):
            stopsproxy = context.make_helper(builder, self.stopstype, stops)
            newval = numba.np.arrayobj._getitem_array_single_int(
                context,
                builder,
                numba.intp,
                self.stopstype,
                stopsproxy,
                builder.sub(partitionid, context.get_constant(numba.intp, 1)),
            )
            builder.store(newval, out)

        return builder.load(out)

    def lower_get_localstop(self, context, builder, stops, partitionid):
        stopsproxy = context.make_helper(builder, self.stopstype, stops)
        return numba.np.arrayobj._getitem_array_single_int(
            context, builder, numba.intp, self.stopstype, stopsproxy, partitionid
        )

    def lower_get_partitionid(
        self, context, builder, pyapi, pylookups, partitionid, viewlength
    ):
        lookup_obj = pyapi.list_getitem(pylookups, partitionid)  # borrowed
        lookup = pyapi.to_native_value(LookupType(), lookup_obj).value
        lookupproxy = context.make_helper(builder, LookupType(), value=lookup)

        viewproxy = context.make_helper(builder, self.toArrayViewType())
        viewproxy.pos = context.get_constant(numba.intp, 0)
        viewproxy.start = context.get_constant(numba.intp, 0)
        viewproxy.stop = viewlength
        viewproxy.arrayptrs = context.make_helper(
            builder, LookupType.arraytype, lookupproxy.arrayptrs
        ).data
        viewproxy.sharedptrs = context.make_helper(
            builder, LookupType.arraytype, lookupproxy.sharedptrs
        ).data
        viewproxy.pylookup = lookup_obj
        return viewproxy._getvalue()


@numba.extending.register_model(PartitionedViewType)
class PartitionedViewModel(numba.core.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ("pylookups", numba.types.pyobject),
            ("partitionid", numba.types.CPointer(numba.intp)),
            ("stops", fe_type.stopstype),
            ("view", numba.types.CPointer(fe_type.toArrayViewType())),
            ("start", numba.intp),
            ("stop", numba.intp),
        ]
        super(PartitionedViewModel, self).__init__(dmm, fe_type, members)


@numba.extending.unbox(PartitionedViewType)
def unbox_PartitionedArray(partviewtype, arrayobj, c):
    partview_obj = c.pyapi.object_getattr_string(arrayobj, "_numbaview")
    out = unbox_PartitionedView(partviewtype, partview_obj, c)
    c.pyapi.decref(partview_obj)
    return out


def unbox_PartitionedView(partviewtype, partview_obj, c):
    lookups_obj = c.pyapi.object_getattr_string(partview_obj, "lookups")
    stops_obj = c.pyapi.object_getattr_string(partview_obj, "stops")
    start_obj = c.pyapi.object_getattr_string(partview_obj, "start")
    stop_obj = c.pyapi.object_getattr_string(partview_obj, "stop")

    proxyout = c.context.make_helper(c.builder, partviewtype)
    proxyout.pylookups = lookups_obj
    partitionid = c.context.get_constant(numba.intp, 0)
    proxyout.partitionid = numba.core.cgutils.alloca_once_value(c.builder, partitionid)
    proxyout.stops = c.pyapi.to_native_value(partviewtype.stopstype, stops_obj).value
    viewlength = partviewtype.lower_get_localstop(
        c.context, c.builder, proxyout.stops, partitionid
    )
    proxyout.view = numba.core.cgutils.alloca_once_value(
        c.builder,
        partviewtype.lower_get_partitionid(
            c.context, c.builder, c.pyapi, proxyout.pylookups, partitionid, viewlength
        ),
    )
    proxyout.start = c.pyapi.number_as_ssize_t(start_obj)
    proxyout.stop = c.pyapi.number_as_ssize_t(stop_obj)

    c.pyapi.decref(lookups_obj)
    c.pyapi.decref(stops_obj)
    c.pyapi.decref(start_obj)
    c.pyapi.decref(stop_obj)

    is_error = numba.core.cgutils.is_not_null(c.builder, c.pyapi.err_occurred())
    return numba.extending.NativeValue(proxyout._getvalue(), is_error)


@numba.extending.box(PartitionedViewType)
def box_PartitionedArray(partviewtype, partviewval, c):
    arrayview_obj = box_PartitionedView(partviewtype, partviewval, c)
    out = c.pyapi.call_method(arrayview_obj, "toarray", ())
    c.pyapi.decref(arrayview_obj)
    return out


def box_PartitionedView(partviewtype, partviewval, c):
    serializable2dict_obj = c.pyapi.unserialize(
        c.pyapi.serialize_object(serializable2dict)
    )
    behavior2_obj = c.pyapi.unserialize(
        c.pyapi.serialize_object(dict2serializable(partviewtype.behavior))
    )
    behavior_obj = c.pyapi.call_function_objargs(
        serializable2dict_obj, (behavior2_obj,)
    )
    PartitionedView_obj = c.pyapi.unserialize(c.pyapi.serialize_object(PartitionedView))
    type_obj = c.pyapi.unserialize(c.pyapi.serialize_object(partviewtype.type))
    fields_obj = c.pyapi.unserialize(c.pyapi.serialize_object(partviewtype.fields))

    proxyin = c.context.make_helper(c.builder, partviewtype, partviewval)
    lookups_obj = proxyin.pylookups
    stops_obj = c.pyapi.from_native_value(
        partviewtype.stopstype, proxyin.stops, c.env_manager
    )
    start_obj = c.pyapi.long_from_ssize_t(proxyin.start)
    stop_obj = c.pyapi.long_from_ssize_t(proxyin.stop)

    out = c.pyapi.call_function_objargs(
        PartitionedView_obj,
        (
            type_obj,
            behavior_obj,
            lookups_obj,
            stops_obj,
            start_obj,
            stop_obj,
            fields_obj,
        ),
    )

    c.pyapi.decref(serializable2dict_obj)
    c.pyapi.decref(behavior2_obj)
    c.pyapi.decref(behavior_obj)
    c.pyapi.decref(PartitionedView_obj)
    c.pyapi.decref(type_obj)
    c.pyapi.decref(fields_obj)
    c.pyapi.decref(stops_obj)
    c.pyapi.decref(start_obj)
    c.pyapi.decref(stop_obj)

    return out


@numba.core.typing.templates.infer_global(operator.getitem)
class type_getitem_partitioned(numba.core.typing.templates.AbstractTemplate):
    def generic(self, args, kwargs):
        if (
            len(args) == 2
            and len(kwargs) == 0
            and isinstance(args[0], PartitionedViewType)
        ):
            partviewtype, wheretype = args

            if isinstance(wheretype, numba.types.Integer):
                arrayviewtype = partviewtype.toArrayViewType()
                rettype = partviewtype.type.getitem_at_check(arrayviewtype)
                return rettype(partviewtype, wheretype)

            elif (
                isinstance(wheretype, numba.types.SliceType) and not wheretype.has_step
            ):
                return partviewtype(partviewtype, wheretype)

            elif isinstance(wheretype, numba.types.StringLiteral):
                rettype = partviewtype.getitem_field(wheretype.literal_value)
                return rettype(partviewtype, wheretype)

            else:
                raise TypeError(
                    "only an integer, start:stop range, or a *constant* "
                    "field name string may be used as ak.Array "
                    "slices in compiled code" + ak._util.exception_suffix(__file__)
                )


@numba.extending.lower_builtin(
    operator.getitem, PartitionedViewType, numba.types.Integer
)
def lower_getitem_at_partitioned(context, builder, sig, args):
    rettype, (partviewtype, wheretype) = sig.return_type, sig.args
    partviewval, whereval = args
    partviewproxy = context.make_helper(builder, partviewtype, partviewval)

    length = builder.sub(partviewproxy.stop, partviewproxy.start)
    regular_atval = numba.core.cgutils.alloca_once_value(builder, whereval)

    with builder.if_then(
        builder.icmp_signed("<", whereval, context.get_constant(numba.intp, 0))
    ):
        builder.store(builder.add(whereval, length), regular_atval)
    atval = builder.load(regular_atval)

    with builder.if_then(
        builder.or_(
            builder.icmp_signed("<", atval, context.get_constant(numba.intp, 0)),
            builder.icmp_signed(">=", atval, length),
        )
    ):
        context.call_conv.return_user_exc(
            builder, ValueError, ("slice index out of bounds",)
        )

    localstart = partviewtype.lower_get_localstart(
        context, builder, partviewproxy.stops, builder.load(partviewproxy.partitionid)
    )
    localstop = partviewtype.lower_get_localstop(
        context, builder, partviewproxy.stops, builder.load(partviewproxy.partitionid)
    )

    with builder.if_then(
        builder.not_(
            builder.and_(
                builder.icmp_signed("<=", localstart, atval),
                builder.icmp_signed(">", atval, localstop),
            )
        ),
        likely=False,
    ):

        searchsorted_sig = numba.intp(partviewtype.stopstype, wheretype)
        searchsorted_args = (partviewproxy.stops, atval)

        def searchsorted_impl(stops, where):
            return ak.nplike.numpy.searchsorted(stops, where, side="right")

        partitionid_val = context.compile_internal(
            builder, searchsorted_impl, searchsorted_sig, searchsorted_args
        )
        builder.store(partitionid_val, partviewproxy.partitionid)

        pyapi = context.get_python_api(builder)
        gil = pyapi.gil_ensure()
        builder.store(
            partviewtype.lower_get_partitionid(
                context,
                builder,
                pyapi,
                partviewproxy.pylookups,
                builder.load(partviewproxy.partitionid),
                builder.sub(localstop, localstart),
            ),
            partviewproxy.view,
        )
        pyapi.gil_release(gil)

    viewtype = partviewtype.toArrayViewType()
    viewval = builder.load(partviewproxy.view)
    viewproxy = context.make_helper(builder, viewtype, value=viewval)

    reallocalstart = partviewtype.lower_get_localstart(
        context, builder, partviewproxy.stops, builder.load(partviewproxy.partitionid)
    )
    subatval = builder.sub(atval, reallocalstart)

    return viewtype.type.lower_getitem_at_check(
        context,
        builder,
        rettype,
        viewtype,
        viewval,
        viewproxy,
        numba.intp,
        subatval,
        False,
        False,
    )


@numba.extending.lower_builtin(
    operator.getitem, PartitionedViewType, numba.types.slice2_type
)
def lower_getitem_range_partitioned(context, builder, sig, args):
    _, (partviewtype, wheretype) = sig.return_type, sig.args
    partviewval, whereval = args

    whereproxy = context.make_helper(builder, wheretype, whereval)
    start = whereproxy.start
    stop = whereproxy.stop

    partviewproxy = context.make_helper(builder, partviewtype, partviewval)
    length = builder.sub(partviewproxy.stop, partviewproxy.start)

    regular_start = numba.core.cgutils.alloca_once_value(builder, start)
    regular_stop = numba.core.cgutils.alloca_once_value(builder, stop)

    with builder.if_then(
        builder.icmp_signed("<", start, context.get_constant(numba.intp, 0))
    ):
        builder.store(builder.add(start, length), regular_start)
    with builder.if_then(
        builder.icmp_signed("<", stop, context.get_constant(numba.intp, 0))
    ):
        builder.store(builder.add(stop, length), regular_stop)

    with builder.if_then(
        builder.icmp_signed(
            "<", builder.load(regular_start), context.get_constant(numba.intp, 0)
        )
    ):
        builder.store(context.get_constant(numba.intp, 0), regular_start)
    with builder.if_then(builder.icmp_signed(">", builder.load(regular_start), length)):
        builder.store(length, regular_start)

    with builder.if_then(
        builder.icmp_signed(
            "<", builder.load(regular_stop), builder.load(regular_start)
        )
    ):
        builder.store(builder.load(regular_start), regular_stop)
    with builder.if_then(builder.icmp_signed(">", builder.load(regular_stop), length)):
        builder.store(length, regular_stop)

    proxyout = context.make_helper(builder, partviewtype)
    proxyout.pylookups = partviewproxy.pylookups
    proxyout.partitionid = numba.core.cgutils.alloca_once_value(
        builder, builder.load(partviewproxy.partitionid)
    )
    proxyout.stops = partviewproxy.stops
    proxyout.view = numba.core.cgutils.alloca_once_value(
        builder, builder.load(partviewproxy.view)
    )
    proxyout.start = builder.load(regular_start)
    proxyout.stop = builder.load(regular_stop)

    if context.enable_nrt:
        context.nrt.incref(builder, partviewtype.stopstype, proxyout.stops)

    return proxyout._getvalue()


@numba.extending.lower_builtin(
    operator.getitem, PartitionedViewType, numba.types.StringLiteral
)
def lower_getitem_field_partitioned(context, builder, sig, args):
    _, (partviewtype, _) = sig.return_type, sig.args
    partviewval, whereval = args

    if context.enable_nrt:
        partviewproxy = context.make_helper(builder, partviewtype, partviewval)
        context.nrt.incref(builder, partviewtype.stopstype, partviewproxy.stops)

    return partviewval


@numba.core.typing.templates.infer_global(len)
class type_len_partitioned(numba.core.typing.templates.AbstractTemplate):
    def generic(self, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], PartitionedViewType)
        ):
            return numba.intp(args[0])


@numba.extending.lower_builtin(len, PartitionedViewType)
def lower_len_partitioned(context, builder, sig, args):
    proxyin = context.make_helper(builder, sig.args[0], args[0])
    return builder.sub(proxyin.stop, proxyin.start)


@numba.core.typing.templates.infer_getattr
class type_getattr_partitioned(numba.core.typing.templates.AttributeTemplate):
    key = PartitionedViewType

    def generic_resolve(self, partviewtype, attr):
        if attr == "ndim":
            return numba.intp
        else:
            return partviewtype.getitem_field(attr)


@numba.extending.lower_getattr_generic(PartitionedViewType)
def lower_getattr_generic_partitioned(
    context, builder, partviewtype, partviewval, attr
):
    if attr == "ndim":
        return context.get_constant(numba.intp, partviewtype.type.ndim)
    elif context.enable_nrt:
        partviewproxy = context.make_helper(builder, partviewtype, partviewval)
        context.nrt.incref(builder, partviewtype.stopstype, partviewproxy.stops)

    return partviewval


class PartitionedIteratorType(numba.types.common.SimpleIteratorType):
    def __init__(self, partviewtype):
        super(PartitionedIteratorType, self).__init__(
            "ak.PartitionedIterator({0})".format(partviewtype.name),
            partviewtype.type.getitem_at_check(partviewtype.toArrayViewType()),
        )
        self.partviewtype = partviewtype


@numba.core.typing.templates.infer
class type_getiter_partitioned(numba.core.typing.templates.AbstractTemplate):
    key = "getiter"

    def generic(self, args, kwargs):
        if (
            len(args) == 1
            and len(kwargs) == 0
            and isinstance(args[0], PartitionedViewType)
        ):
            return PartitionedIteratorType(args[0])(args[0])


@numba.core.datamodel.registry.register_default(PartitionedIteratorType)
class PartitionedIteratorModel(numba.core.datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ("partview", fe_type.partviewtype),
            ("length", numba.intp),
            ("at", numba.types.EphemeralPointer(numba.intp)),
        ]
        super(PartitionedIteratorModel, self).__init__(dmm, fe_type, members)


@numba.extending.lower_builtin("getiter", PartitionedViewType)
def lower_getiter_partitioned(context, builder, sig, args):
    rettype, (partviewtype,) = sig.return_type, sig.args
    (partviewval,) = args
    partviewproxy = context.make_helper(builder, partviewtype, partviewval)

    partitionid = context.get_constant(numba.intp, 0)
    viewlength = partviewtype.lower_get_localstop(
        context, builder, partviewproxy.stops, partitionid
    )

    partoutproxy = context.make_helper(builder, partviewtype)
    partoutproxy.pylookups = partviewproxy.pylookups
    partoutproxy.partitionid = numba.core.cgutils.alloca_once_value(
        builder, partitionid
    )
    partoutproxy.stops = partviewproxy.stops

    pyapi = context.get_python_api(builder)
    gil = pyapi.gil_ensure()
    partoutproxy.view = numba.core.cgutils.alloca_once_value(
        builder,
        partviewtype.lower_get_partitionid(
            context, builder, pyapi, partviewproxy.pylookups, partitionid, viewlength
        ),
    )
    pyapi.gil_release(gil)

    partoutproxy.start = partviewproxy.start
    partoutproxy.stop = partviewproxy.stop

    proxyout = context.make_helper(builder, rettype)
    proxyout.partview = partoutproxy._getvalue()
    proxyout.length = builder.sub(partviewproxy.stop, partviewproxy.start)
    proxyout.at = numba.core.cgutils.alloca_once_value(
        builder, context.get_constant(numba.intp, 0)
    )

    if context.enable_nrt:
        context.nrt.incref(builder, partviewtype.stopstype, partoutproxy.stops)

    return numba.core.imputils.impl_ret_new_ref(
        context, builder, rettype, proxyout._getvalue()
    )


@numba.extending.lower_builtin("iternext", PartitionedIteratorType)
@numba.core.imputils.iternext_impl(numba.core.imputils.RefType.BORROWED)
def lower_iternext_partitioned(context, builder, sig, args, result):
    (itertype,) = sig.args
    (iterval,) = args

    proxyin = context.make_helper(builder, itertype, iterval)
    partviewproxy = context.make_helper(
        builder, itertype.partviewtype, proxyin.partview
    )
    at = builder.load(proxyin.at)

    is_valid = builder.icmp_signed("<", at, proxyin.length)
    result.set_valid(is_valid)

    with builder.if_then(is_valid, likely=True):
        maybestop = itertype.partviewtype.lower_get_localstop(
            context,
            builder,
            partviewproxy.stops,
            builder.load(partviewproxy.partitionid),
        )

        with builder.if_then(builder.icmp_signed("==", at, maybestop)):
            builder.store(
                builder.add(
                    builder.load(partviewproxy.partitionid),
                    context.get_constant(numba.intp, 1),
                ),
                partviewproxy.partitionid,
            )

            localstart = itertype.partviewtype.lower_get_localstart(
                context,
                builder,
                partviewproxy.stops,
                builder.load(partviewproxy.partitionid),
            )
            localstop = itertype.partviewtype.lower_get_localstop(
                context,
                builder,
                partviewproxy.stops,
                builder.load(partviewproxy.partitionid),
            )

            pyapi = context.get_python_api(builder)
            gil = pyapi.gil_ensure()
            builder.store(
                itertype.partviewtype.lower_get_partitionid(
                    context,
                    builder,
                    pyapi,
                    partviewproxy.pylookups,
                    builder.load(partviewproxy.partitionid),
                    builder.sub(localstop, localstart),
                ),
                partviewproxy.view,
            )
            pyapi.gil_release(gil)

        realstart = itertype.partviewtype.lower_get_localstart(
            context,
            builder,
            partviewproxy.stops,
            builder.load(partviewproxy.partitionid),
        )

        outview = builder.load(partviewproxy.view)
        outviewtype = itertype.partviewtype.toArrayViewType()
        outviewproxy = context.make_helper(builder, outviewtype, outview)

        result.yield_(
            itertype.partviewtype.type.lower_getitem_at_check(
                context,
                builder,
                itertype.partviewtype.type.getitem_at_check(outviewtype),
                outviewtype,
                outview,
                outviewproxy,
                numba.intp,
                builder.sub(at, realstart),
                False,
                False,
            )
        )

        nextat = numba.core.cgutils.increment_index(builder, at)
        builder.store(nextat, proxyin.at)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


class Point(ak.Record):
    def distance(self, other):
        return np.sqrt((self.x - other.x) ** 2 + (self.y - other.y) ** 2)


def test():
    behavior = {"point": Point}
    rec = ak.Record({"x": 1, "y": 1.1}, with_name="point", behavior=behavior)
    assert isinstance(rec, Point)
    assert hasattr(rec, "distance")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

pytest.importorskip("numba")


def test_numpy_rhs():
    transform = ak.Array(
        [
            [0, 1, 0],
            [4, 0, 0],
            [0, 0, 2],
        ]
    )

    vector = ak.Array(np.r_[4, 5, 6].reshape(3, 1))

    result = np.matmul(transform, vector)

    assert result.tolist() == [[5], [16], [12]]


def test_numpy_lhs():
    transform = ak.Array(
        [
            [0, 4, 0],
            [1, 0, 0],
            [0, 0, 2],
        ]
    )

    vector = ak.Array(np.r_[4, 5, 6].reshape(1, 3))

    result = np.matmul(vector, transform)

    assert result.tolist() == [[5, 16, 12]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_pandas_style_constructor():
    a = ak.Array({"x": [1, 2, 3], "y": [[1, 2, 3], [], [4, 5]]})
    assert ak.to_list(a) == [
        {"x": 1, "y": [1, 2, 3]},
        {"x": 2, "y": []},
        {"x": 3, "y": [4, 5]},
    ]


pyarrow = pytest.importorskip("pyarrow")


def test_pyarrow_constructor():
    a = ak.Array(pyarrow.array([[1.1, 2.2, 3.3], [], [4.4, 5.5]]))
    assert ak.to_list(a) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import ctypes

import numpy

import awkward as ak


def of(*arrays):
    libs = set()
    for array in arrays:
        nplike = getattr(array, "nplike", None)
        if isinstance(nplike, NumpyLike):
            libs.add(nplike)
        elif isinstance(array, numpy.ndarray):
            ptr_lib = "cpu"
        elif (
            type(array).__module__.startswith("cupy.")
            and type(array).__name__ == "ndarray"
        ):
            ptr_lib = "cuda"
        else:
            ptr_lib = ak.operations.convert.kernels(array)
            if ptr_lib is None:
                pass
            elif ptr_lib == "cpu":
                libs.add("cpu")
            elif ptr_lib == "cuda":
                libs.add("cuda")
            else:
                raise ValueError(
                    """structure mixes 'cpu' and 'cuda' buffers; use one of

    ak.to_kernels(array, 'cpu')
    ak.to_kernels(array, 'cuda')

to obtain an unmixed array in main memory or the GPU(s)."""
                    + ak._util.exception_suffix(__file__)
                )

    if libs == set() or libs == set(["cpu"]):
        return Numpy.instance()
    elif libs == set(["cuda"]):
        return Cupy.instance()
    elif len(libs) == 1:
        return next(iter(libs))
    else:
        raise ValueError(
            """attempting to use both a 'cpu' array and a 'cuda' array in the """
            """same operation; use one of

    ak.to_kernels(array, 'cpu')
    ak.to_kernels(array, 'cuda')

to move one or the other to main memory or the GPU(s)."""
            + ak._util.exception_suffix(__file__)
        )


class Singleton(object):
    _instance = None

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance


class NumpyMetadata(Singleton):
    bool_ = numpy.bool_
    int8 = numpy.int8
    int16 = numpy.int16
    int32 = numpy.int32
    int64 = numpy.int64
    uint8 = numpy.uint8
    uint16 = numpy.uint16
    uint32 = numpy.uint32
    uint64 = numpy.uint64
    float32 = numpy.float32
    float64 = numpy.float64
    complex64 = numpy.complex64
    complex128 = numpy.complex128
    str_ = numpy.str_
    bytes_ = numpy.bytes_

    intp = numpy.intp
    integer = numpy.integer
    signedinteger = numpy.signedinteger
    unsignedinteger = numpy.unsignedinteger
    floating = numpy.floating
    number = numpy.number
    object_ = numpy.object_
    generic = numpy.generic

    dtype = numpy.dtype
    ufunc = numpy.ufunc
    iinfo = numpy.iinfo
    errstate = numpy.errstate
    newaxis = numpy.newaxis

    ndarray = numpy.ndarray

    nan = numpy.nan
    inf = numpy.inf

    nat = numpy.datetime64("NaT")
    datetime_data = numpy.datetime_data
    issubdtype = numpy.issubdtype

    AxisError = numpy.AxisError


if hasattr(numpy, "float16"):
    NumpyMetadata.float16 = numpy.float16

if hasattr(numpy, "float128"):
    NumpyMetadata.float128 = numpy.float128

if hasattr(numpy, "complex256"):
    NumpyMetadata.complex256 = numpy.complex256

if hasattr(numpy, "datetime64"):
    NumpyMetadata.datetime64 = numpy.datetime64

if hasattr(numpy, "timedelta64"):
    NumpyMetadata.timedelta64 = numpy.timedelta64

NumpyMetadata.all_complex = tuple(
    x for x in dir(NumpyMetadata) if x.startswith("complex")
)


class NumpyLike(Singleton):
    known_shape = True
    known_dtype = True

    ############################ array creation

    def array(self, *args, **kwargs):
        # data[, dtype=[, copy=]]
        return self._module.array(*args, **kwargs)

    def asarray(self, *args, **kwargs):
        # array[, dtype=][, order=]
        return self._module.asarray(*args, **kwargs)

    def ascontiguousarray(self, *args, **kwargs):
        # array[, dtype=]
        return self._module.ascontiguousarray(*args, **kwargs)

    def isscalar(self, *args, **kwargs):
        return self._module.isscalar(*args, **kwargs)

    def frombuffer(self, *args, **kwargs):
        # array[, dtype=]
        return self._module.frombuffer(*args, **kwargs)

    def zeros(self, *args, **kwargs):
        # shape/len[, dtype=]
        return self._module.zeros(*args, **kwargs)

    def ones(self, *args, **kwargs):
        # shape/len[, dtype=]
        return self._module.ones(*args, **kwargs)

    def empty(self, *args, **kwargs):
        # shape/len[, dtype=]
        return self._module.empty(*args, **kwargs)

    def full(self, *args, **kwargs):
        # shape/len, value[, dtype=]
        return self._module.full(*args, **kwargs)

    def zeros_like(self, *args, **kwargs):
        # array
        return self._module.zeros_like(*args, **kwargs)

    def ones_like(self, *args, **kwargs):
        # array
        return self._module.ones_like(*args, **kwargs)

    def full_like(self, *args, **kwargs):
        # array, fill_value
        return self._module.full_like(*args, **kwargs)

    def arange(self, *args, **kwargs):
        # stop[, dtype=]
        # start, stop[, dtype=]
        # start, stop, step[, dtype=]
        return self._module.arange(*args, **kwargs)

    def meshgrid(self, *args, **kwargs):
        # *arrays, indexing="ij"
        return self._module.meshgrid(*args, **kwargs)

    ############################ testing

    def shape(self, *args, **kwargs):
        # array
        return self._module.shape(*args, **kwargs)

    def array_equal(self, *args, **kwargs):
        # array1, array2
        return self._module.array_equal(*args, **kwargs)

    def size(self, *args, **kwargs):
        # array
        return self._module.size(*args, **kwargs)

    def searchsorted(self, *args, **kwargs):
        # haystack, needle, side="right"
        return self._module.searchsorted(*args, **kwargs)

    def argsort(self, *args, **kwargs):
        # array
        return self._module.argsort(*args, **kwargs)

    ############################ manipulation

    def broadcast_arrays(self, *args, **kwargs):
        # array1[, array2[, ...]]
        return self._module.broadcast_arrays(*args, **kwargs)

    def add(self, *args, **kwargs):
        # array1, array2[, out=]
        return self._module.add(*args, **kwargs)

    def cumsum(self, *args, **kwargs):
        # arrays[, out=]
        return self._module.cumsum(*args, **kwargs)

    def cumprod(self, *args, **kwargs):
        # arrays[, out=]
        return self._module.cumprod(*args, **kwargs)

    def nonzero(self, *args, **kwargs):
        # array
        return self._module.nonzero(*args, **kwargs)

    def unique(self, *args, **kwargs):
        # array
        return self._module.unique(*args, **kwargs)

    def concatenate(self, *args, **kwargs):
        # arrays
        return self._module.concatenate(*args, **kwargs)

    def repeat(self, *args, **kwargs):
        # array, int
        # array1, array2
        return self._module.repeat(*args, **kwargs)

    def stack(self, *args, **kwargs):
        # arrays
        return self._module.stack(*args, **kwargs)

    def vstack(self, *args, **kwargs):
        # arrays
        return self._module.vstack(*args, **kwargs)

    def packbits(self, *args, **kwargs):
        # array
        return self._module.packbits(*args, **kwargs)

    def unpackbits(self, *args, **kwargs):
        # array
        return self._module.unpackbits(*args, **kwargs)

    def atleast_1d(self, *args, **kwargs):
        # *arrays
        return self._module.atleast_1d(*args, **kwargs)

    def broadcast_to(self, *args, **kwargs):
        # array, shape
        return self._module.broadcast_to(*args, **kwargs)

    def append(self, *args, **kwargs):
        # array, element
        return self._module.append(*args, **kwargs)

    def where(self, *args, **kwargs):
        # array, element
        return self._module.where(*args, **kwargs)

    ############################ ufuncs

    def sqrt(self, *args, **kwargs):
        # array
        return self._module.sqrt(*args, **kwargs)

    def exp(self, *args, **kwargs):
        # array
        return self._module.exp(*args, **kwargs)

    def true_divide(self, *args, **kwargs):
        # array1, array2
        return self._module.true_divide(*args, **kwargs)

    def bitwise_or(self, *args, **kwargs):
        # array1, array2[, out=output]
        return self._module.bitwise_or(*args, **kwargs)

    def logical_and(self, *args, **kwargs):
        # array1, array2
        return self._module.logical_and(*args, **kwargs)

    def equal(self, *args, **kwargs):
        # array1, array2
        return self._module.equal(*args, **kwargs)

    def ceil(self, *args, **kwargs):
        # array
        return self._module.ceil(*args, **kwargs)

    ############################ almost-ufuncs

    def nan_to_num(self, *args, **kwargs):
        # array, copy=True, nan=0.0, posinf=None, neginf=None
        return self._module.nan_to_num(*args, **kwargs)

    def isclose(self, *args, **kwargs):
        # a, b, rtol=1e-05, atol=1e-08, equal_nan=False
        return self._module.isclose(*args, **kwargs)

    ############################ reducers

    def all(self, *args, **kwargs):
        # array
        return self._module.all(*args, **kwargs)

    def any(self, *args, **kwargs):
        # array
        kwargs.pop("prefer", None)
        return self._module.any(*args, **kwargs)

    def count_nonzero(self, *args, **kwargs):
        # array
        return self._module.count_nonzero(*args, **kwargs)

    def sum(self, *args, **kwargs):
        # array
        return self._module.sum(*args, **kwargs)

    def prod(self, *args, **kwargs):
        # array
        return self._module.prod(*args, **kwargs)

    def min(self, *args, **kwargs):
        # array
        return self._module.min(*args, **kwargs)

    def max(self, *args, **kwargs):
        # array
        return self._module.max(*args, **kwargs)

    def argmin(self, *args, **kwargs):
        # array[, axis=]
        return self._module.argmin(*args, **kwargs)

    def argmax(self, *args, **kwargs):
        # array[, axis=]
        return self._module.argmax(*args, **kwargs)

    def array_str(self, *args, **kwargs):
        # array, max_line_width, precision=None, suppress_small=None
        return self._module.array_str(*args, **kwargs)

    def datetime_as_string(self, *args, **kwargs):
        return self._module.datetime_as_string(*args, **kwargs)


class NumpyKernel(object):
    def __init__(self, kernel, name_and_types):
        self._kernel = kernel
        self._name_and_types = name_and_types

    def __repr__(self):
        return "<{0} {1}{2}>".format(
            type(self).__name__,
            self._name_and_types[0],
            "".join(", " + str(numpy.dtype(x)) for x in self._name_and_types[1:]),
        )

    @staticmethod
    def _cast(x, t):
        if issubclass(t, ctypes._Pointer):
            if isinstance(x, numpy.ndarray):
                return ctypes.cast(x.ctypes.data, t)
            else:
                return ctypes.cast(x, t)
        else:
            return x

    def __call__(self, *args):
        assert len(args) == len(self._kernel.argtypes)
        return self._kernel(
            *(self._cast(x, t) for x, t in zip(args, self._kernel.argtypes))
        )


class Numpy(NumpyLike):
    def to_rectilinear(self, array, *args, **kwargs):
        return ak.operations.convert.to_numpy(array, *args, **kwargs)

    def __getitem__(self, name_and_types):
        return NumpyKernel(ak._cpu_kernels.kernel[name_and_types], name_and_types)

    def __init__(self):
        self._module = numpy

    @property
    def ma(self):
        return self._module.ma

    @property
    def char(self):
        return self._module.char

    @property
    def ndarray(self):
        return self._module.ndarray


class Cupy(NumpyLike):
    def to_rectilinear(self, array, *args, **kwargs):
        return ak.operations.convert.to_cupy(array, *args, **kwargs)

    def __getitem__(self, name_and_types):
        raise NotImplementedError("no CUDA in v2 yet")

    def __init__(self):
        try:
            import cupy
        except ImportError:
            raise ImportError(
                """to use CUDA arrays in Python, install the 'cupy' package with:

    pip install cupy --upgrade

or

    conda install cupy"""
            )
        self._module = cupy

    @property
    def ma(self):
        raise ValueError(
            "CUDA arrays cannot have missing values until CuPy implements "
            "numpy.ma.MaskedArray" + ak._util.exception_suffix(__file__)
        )

    @property
    def char(self):
        raise ValueError(
            "CUDA arrays cannot do string manipulations until CuPy implements "
            "numpy.char" + ak._util.exception_suffix(__file__)
        )

    @property
    def ndarray(self):
        return self._module.ndarray

    def asarray(self, array, dtype=None):
        if isinstance(
            array,
            (
                ak.highlevel.Array,
                ak.highlevel.Record,
                ak.layout.Content,
                ak.layout.Record,
            ),
        ):
            out = ak.operations.convert.to_cupy(array)
            if dtype is not None and out.dtype != dtype:
                return self._module.asarray(out, dtype=dtype)
            else:
                return out
        else:
            return self._module.asarray(array, dtype=dtype)

    def ascontiguousarray(self, array, dtype=None):
        if isinstance(
            array,
            (
                ak.highlevel.Array,
                ak.highlevel.Record,
                ak.layout.Content,
                ak.layout.Record,
            ),
        ):
            out = ak.operations.convert.to_cupy(array)
            if dtype is not None and out.dtype != dtype:
                return self._module.ascontiguousarray(out, dtype=dtype)
            else:
                return out
        else:
            return self._module.ascontiguousarray(array, dtype=dtype)

    def zeros(self, *args, **kwargs):
        return self._module.zeros(*args, **kwargs)

    def frombuffer(self, *args, **kwargs):
        np_array = numpy.frombuffer(*args, **kwargs)
        return self._module.array(np_array)

    def array_equal(self, array1, array2):
        # CuPy issue?
        if array1.shape != array2.shape:
            return False
        else:
            return self._module.all(array1 - array2 == 0)

    def repeat(self, array, repeats):
        # https://github.com/cupy/cupy/issues/3849
        if isinstance(repeats, self._module.ndarray):
            all_stops = self._module.cumsum(repeats)
            parents = self._module.zeros(all_stops[-1].item(), dtype=int)
            stops, stop_counts = self._module.unique(all_stops[:-1], return_counts=True)
            parents[stops] = stop_counts
            self._module.cumsum(parents, out=parents)
            return array[parents]
        else:
            return self._module.repeat(array, repeats)

    def nan_to_num(self, array, copy=True, nan=0.0, posinf=None, neginf=None):
        # https://github.com/cupy/cupy/issues/4867
        if copy:
            array = self._module.copy(array)
        if posinf is None:
            if array.dtype.kind == "f":
                posinf = numpy.finfo(array.dtype.type).max
            else:
                posinf = numpy.iinfo(array.dtype.type).max
        if neginf is None:
            if array.dtype.kind == "f":
                neginf = numpy.finfo(array.dtype.type).min
            else:
                neginf = numpy.iinfo(array.dtype.type).min

        array[self._module.isnan(array)] = nan
        array[self._module.isinf(array) & (array > 0)] = posinf
        array[self._module.isinf(array) & (array < 0)] = neginf
        return array

    # For all reducers: https://github.com/cupy/cupy/issues/3819

    def all(self, array, axis=None, **kwargs):
        kwargs.pop("prefer", None)
        out = self._module.all(array, axis=axis)
        if axis is None and isinstance(out, self._module.ndarray):
            return out.item()
        else:
            return out

    def any(self, array, axis=None, **kwargs):
        kwargs.pop("prefer", None)
        out = self._module.any(array, axis=axis)
        if axis is None and isinstance(out, self._module.ndarray):
            return out.item()
        else:
            return out

    def count_nonzero(self, array, axis=None):
        out = self._module.count_nonzero(array, axis=axis)
        if axis is None and isinstance(out, self._module.ndarray):
            return out.item()
        else:
            return out

    def sum(self, array, axis=None):
        out = self._module.sum(array, axis=axis)
        if axis is None and isinstance(out, self._module.ndarray):
            return out.item()
        else:
            return out

    def prod(self, array, axis=None):
        out = self._module.prod(array, axis=axis)
        if axis is None and isinstance(out, self._module.ndarray):
            return out.item()
        else:
            return out

    def min(self, array, axis=None):
        out = self._module.min(array, axis=axis)
        if axis is None and isinstance(out, self._module.ndarray):
            return out.item()
        else:
            return out

    def max(self, array, axis=None):
        out = self._module.max(array, axis=axis)
        if axis is None and isinstance(out, self._module.ndarray):
            return out.item()
        else:
            return out

    def argmin(self, array, axis=None):
        out = self._module.argmin(array, axis=axis)
        if axis is None and isinstance(out, self._module.ndarray):
            return out.item()
        else:
            return out

    def argmax(self, array, axis=None):
        out = self._module.argmax(array, axis=axis)
        if axis is None and isinstance(out, self._module.ndarray):
            return out.item()
        else:
            return out

    def array_str(
        self, array, max_line_width=None, precision=None, suppress_small=None
    ):
        # array, max_line_width, precision=None, suppress_small=None
        return self._module.array_str(array, max_line_width, precision, suppress_small)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

import datetime


def test_date_time():

    numpy_array = np.array(
        ["2020-07-27T10:41:11", "2019-01-01", "2020-01-01"], "datetime64[s]"
    )

    array = ak.Array(numpy_array)
    assert str(array.type) == "3 * datetime64"
    assert array.tolist() == [
        np.datetime64("2020-07-27T10:41:11"),
        np.datetime64("2019-01-01T00:00:00"),
        np.datetime64("2020-01-01T00:00:00"),
    ]
    for i in range(len(array)):
        assert ak.to_numpy(array)[i] == numpy_array[i]

    date_time = np.datetime64("2020-07-27T10:41:11.200000011", "us")
    array1 = ak.Array(np.array(["2020-07-27T10:41:11.200000011"], "datetime64[us]"))
    assert np.datetime64(array1[0], "us") == date_time

    assert ak.to_list(ak.from_iter(array1)) == [
        np.datetime64("2020-07-27T10:41:11.200000")
    ]

    assert ak.max(array) == numpy_array[0]
    assert ak.min(array) == numpy_array[1]


def test_time_delta():

    numpy_array = np.array(["41", "1", "20"], "timedelta64[D]")

    array = ak.Array(numpy_array)
    assert str(array.type) == "3 * timedelta64"
    assert array.tolist() == [
        np.timedelta64("41", "D"),
        np.timedelta64("1", "D"),
        np.timedelta64("20", "D"),
    ]
    for i in range(len(array)):
        assert ak.to_numpy(array)[i] == numpy_array[i]


def test_datetime64_ArrayBuilder():
    builder = ak.layout.ArrayBuilder()
    dt = np.datetime64("2020-03-27T10:41:12", "25us")
    dt1 = np.datetime64("2020-03-27T10:41", "15s")
    dt2 = np.datetime64("2020-05")
    builder.datetime(dt1)
    builder.datetime("2020-03-27T10:41:11")
    builder.datetime(dt)
    builder.datetime("2021-03-27")
    builder.datetime("2020-03-27T10:41:13")
    builder.datetime(dt2)
    builder.datetime("2020-05-01T00:00:00.000000")
    builder.datetime("2020-07-27T10:41:11.200000")

    assert ak.to_list(builder.snapshot()) == [
        datetime.datetime(2020, 3, 27, 10, 41),
        datetime.datetime(2020, 3, 27, 10, 41, 11),
        datetime.datetime(2020, 3, 27, 10, 41, 12),
        datetime.datetime(2021, 3, 27, 0, 0),
        datetime.datetime(2020, 3, 27, 10, 41, 13),
        datetime.datetime(2020, 5, 1, 20, 56, 24),
        datetime.datetime(2020, 5, 1, 0, 0),
        datetime.datetime(2020, 7, 27, 10, 41, 11, 200000),
    ]
    # FIXME?: the above works with both 2.7 and 3, the comparison beneath works in Python 3, but not in 2.7
    # [
    #     np.datetime64("2020-03-27T10:41:00.000000"),
    #     np.datetime64("2020-03-27T10:41:11.000000"),
    #     np.datetime64("2020-03-27T10:41:12.000000"),
    #     np.datetime64("2021-03-27T00:00:00.000000"),
    #     np.datetime64("2020-03-27T10:41:13.000000"),
    #     np.datetime64("2020-05-01T20:56:24.000000"),  # FIXME? precision
    #     np.datetime64("2020-05-01T00:00:00.000000"),
    #     np.datetime64("2020-07-27T10:41:11.200000"),
    # ]


def test_highlevel_datetime64_ArrayBuilder():
    builder = ak.ArrayBuilder()
    dt = np.datetime64("2020-03-27T10:41:12", "25us")
    dt1 = np.datetime64("2020-03-27T10:41", "15s")
    dt2 = np.datetime64("2020-05")
    builder.datetime(dt1)
    builder.datetime("2020-03-27T10:41:11")
    builder.datetime(dt)
    builder.datetime("2021-03-27")
    builder.datetime("2020-03-27T10:41:13")
    builder.timedelta(np.timedelta64(5, "s"))
    builder.datetime(dt2)
    builder.datetime("2020-05-01T00:00:00.000000")
    builder.datetime("2020-07-27T10:41:11.200000")
    builder.integer(1)
    builder.timedelta(np.timedelta64(5, "s"))

    assert ak.to_list(builder.snapshot()) == [
        np.datetime64("2020-03-27T10:41:00.000000"),
        np.datetime64("2020-03-27T10:41:11.000000"),
        np.datetime64("2020-03-27T10:41:12.000000"),
        np.datetime64("2021-03-27T00:00:00.000000"),
        np.datetime64("2020-03-27T10:41:13.000000"),
        np.timedelta64(5, "s"),
        np.datetime64("2020-05-01T20:56:24.000000"),
        np.datetime64("2020-05-01T00:00:00.000000"),
        np.datetime64("2020-07-27T10:41:11.200000"),
        1,
        np.timedelta64(5, "s"),
    ]


def test_timedelta64_ArrayBuilder():
    builder = ak.layout.ArrayBuilder()
    builder.timedelta(np.timedelta64(5, "W"))
    builder.timedelta(np.timedelta64(5, "D"))
    builder.timedelta(np.timedelta64(5, "s"))

    assert ak.to_list(builder.snapshot()) == [
        datetime.timedelta(weeks=5),
        datetime.timedelta(5),
        datetime.timedelta(0, 5),
    ]


@pytest.mark.skipif(
    ak._util.py27,
    reason="Python 2.7 to_list returns datetime.timedelta see the test above",
)
def test_timedelta64_ArrayBuilder_py3():
    builder = ak.layout.ArrayBuilder()
    builder.timedelta(np.timedelta64(5, "W"))
    builder.timedelta(np.timedelta64(5, "D"))
    builder.timedelta(np.timedelta64(5, "s"))

    assert ak.to_list(builder.snapshot()) == [
        np.timedelta64(5 * 7 * 24 * 60 * 60, "s"),
        np.timedelta64(5 * 24 * 60 * 60, "s"),
        np.timedelta64(5, "s"),
    ]


def test_highlevel_timedelta64_ArrayBuilder():
    builder = ak.ArrayBuilder()
    builder.timedelta(np.timedelta64(5, "W"))
    builder.timedelta(np.timedelta64(5, "D"))
    builder.timedelta(np.timedelta64(5, "s"))
    builder.integer(1)
    builder.datetime("2020-05-01T00:00:00.000000")

    assert ak.to_list(builder.snapshot()) == [
        np.timedelta64(5 * 7 * 24 * 60 * 60, "s"),
        np.timedelta64(5 * 24 * 60 * 60, "s"),
        np.timedelta64(5, "s"),
        1,
        np.datetime64("2020-05-01T00:00:00.000000"),
    ]


def test_count():
    array = ak.Array(
        [
            [
                [np.datetime64("2022"), np.datetime64("2023"), np.datetime64("2025")],
                [],
                [np.datetime64("2027"), np.datetime64("2011")],
                [np.datetime64("2013")],
            ],
            [],
            [[np.datetime64("2017"), np.datetime64("2019")], [np.datetime64("2023")]],
        ],
        check_valid=True,
    )
    assert ak.count(array) == 9
    assert ak.to_list(ak.count(array, axis=-1)) == [[3, 0, 2, 1], [], [2, 1]]
    assert ak.to_list(ak.count(array, axis=2)) == [[3, 0, 2, 1], [], [2, 1]]
    assert ak.to_list(ak.count(array, axis=-1, keepdims=True)) == [
        [[3], [0], [2], [1]],
        [],
        [[2], [1]],
    ]
    assert ak.to_list(ak.count(array, axis=-2)) == [[3, 2, 1], [], [2, 1]]
    assert ak.to_list(ak.count(array, axis=1)) == [[3, 2, 1], [], [2, 1]]
    assert ak.to_list(ak.count(array, axis=-2, keepdims=True)) == [
        [[3, 2, 1]],
        [[]],
        [[2, 1]],
    ]


def test_count_nonzero():
    array = ak.Array(
        [
            [
                [np.datetime64("2022"), np.datetime64("2023"), np.datetime64("2025")],
                [],
                [np.datetime64("2027"), np.datetime64("2011")],
                [np.datetime64("2013")],
            ],
            [],
            [[np.datetime64("2017"), np.datetime64("2019")], [np.datetime64("2023")]],
        ],
        check_valid=True,
    )

    assert ak.count_nonzero(array) == 9
    assert ak.to_list(ak.count_nonzero(array, axis=-1)) == [[3, 0, 2, 1], [], [2, 1]]
    assert ak.to_list(ak.count_nonzero(array, axis=-2)) == [[3, 2, 1], [], [2, 1]]

    assert ak.to_list(ak.all(array, axis=-1)) == [
        [True, True, True, True],
        [],
        [True, True],
    ]
    assert ak.to_list(ak.all(array, axis=-2)) == [[True, True, True], [], [True, True]]


def test_argmin_argmax():
    array = ak.Array(
        [
            [
                [np.datetime64("2022"), np.datetime64("2023"), np.datetime64("2025")],
                [],
                [np.datetime64("2027"), np.datetime64("2011")],
                [np.datetime64("2013")],
            ],
            [],
            [[np.datetime64("2017"), np.datetime64("2019")], [np.datetime64("2023")]],
        ],
        check_valid=True,
    )
    assert ak.argmin(array) == 4
    assert ak.argmax(array) == 3
    assert ak.to_list(ak.argmin(array, axis=0)) == [[1, 1, 0], [1], [0, 0], [0]]
    assert ak.to_list(ak.argmax(array, axis=0)) == [[0, 0, 0], [1], [0, 0], [0]]
    assert ak.to_list(ak.argmin(array, axis=1)) == [[3, 2, 0], [], [0, 0]]
    assert ak.to_list(ak.argmax(array, axis=1)) == [[2, 0, 0], [], [1, 0]]

    array = ak.from_iter(
        [
            [
                [
                    np.datetime64("2021-01-20"),
                    np.datetime64("2021-01-10"),
                    np.datetime64("2021-01-30"),
                ]
            ],
            [[]],
            [None, None, None],
            [
                [
                    np.datetime64("2021-01-14"),
                    np.datetime64("2021-01-15"),
                    np.datetime64("2021-01-16"),
                ]
            ],
        ],
        highlevel=False,
    )
    assert ak.to_list(array.argmin(axis=2)) == [[1], [None], [None, None, None], [0]]


def test_any_all():
    array = ak.Array(
        [
            [
                [np.datetime64("2022"), np.datetime64("2023"), np.datetime64("2025")],
                [],
                [np.datetime64("2027"), np.datetime64("2011")],
                [np.datetime64("2013")],
            ],
            [],
            [[np.datetime64("2017"), np.datetime64("2019")], [np.datetime64("2023")]],
        ],
        check_valid=True,
    )

    assert ak.to_list(ak.any(array, axis=-1)) == [
        [True, False, True, True],
        [],
        [True, True],
    ]
    assert ak.to_list(ak.any(array, axis=-2)) == [[True, True, True], [], [True, True]]


def test_prod():
    array = ak.Array(
        np.array(["2020-07-27T10:41:11", "2019-01-01", "2020-01-01"], "datetime64[s]")
    )
    with pytest.raises(ValueError):
        ak.prod(array, axis=-1)


def test_min_max():
    array = ak.Array(
        [
            [
                np.datetime64("2020-03-27T10:41:11"),
                np.datetime64("2020-01-27T10:41:11"),
                np.datetime64("2020-05"),
                np.datetime64("2020-01-27T10:41:11"),
                np.datetime64("2020-04-27T10:41:11"),
            ],
            [
                np.datetime64("2020-04-27"),
                np.datetime64("2020-02-27T10:41:11"),
                np.datetime64("2020-01-27T10:41:11"),
                np.datetime64("2020-06-27T10:41:11"),
            ],
            [
                np.datetime64("2020-02-27T10:41:11"),
                np.datetime64("2020-03-27T10:41:11"),
                np.datetime64("2020-01-27T10:41:11"),
            ],
        ]
    )
    assert ak.to_list(array) == [
        [
            np.datetime64("2020-03-27T10:41:11"),
            np.datetime64("2020-01-27T10:41:11"),
            np.datetime64("2020-05-01T20:56:24"),
            np.datetime64("2020-01-27T10:41:11"),
            np.datetime64("2020-04-27T10:41:11"),
        ],
        [
            np.datetime64("2020-04-27T00:00:00"),
            np.datetime64("2020-02-27T10:41:11"),
            np.datetime64("2020-01-27T10:41:11"),
            np.datetime64("2020-06-27T10:41:11"),
        ],
        [
            np.datetime64("2020-02-27T10:41:11"),
            np.datetime64("2020-03-27T10:41:11"),
            np.datetime64("2020-01-27T10:41:11"),
        ],
    ]
    assert ak.min(array) == np.datetime64("2020-01-27T10:41:11")
    assert ak.max(array) == np.datetime64("2020-06-27T10:41:11")
    assert ak.to_list(ak.min(array, axis=0)) == [
        np.datetime64("2020-02-27T10:41:11"),
        np.datetime64("2020-01-27T10:41:11"),
        np.datetime64("2020-01-27T10:41:11"),
        np.datetime64("2020-01-27T10:41:11"),
        np.datetime64("2020-04-27T10:41:11"),
    ]
    assert ak.to_list(ak.max(array, axis=0)) == [
        np.datetime64("2020-04-27T00:00:00"),
        np.datetime64("2020-03-27T10:41:11"),
        np.datetime64("2020-05-01T20:56:24"),
        np.datetime64("2020-06-27T10:41:11"),
        np.datetime64("2020-04-27T10:41:11"),
    ]
    assert ak.to_list(ak.min(array, axis=1)) == [
        np.datetime64("2020-01-27T10:41:11"),
        np.datetime64("2020-01-27T10:41:11"),
        np.datetime64("2020-01-27T10:41:11"),
    ]
    assert ak.to_list(ak.max(array, axis=1)) == [
        np.datetime64("2020-05-01T20:56:24"),
        np.datetime64("2020-06-27T10:41:11"),
        np.datetime64("2020-03-27T10:41:11"),
    ]


def test_date_time_units():
    array1 = np.array(
        ["2020-07-27T10:41:11", "2019-01-01", "2020-01-01"], "datetime64[s]"
    )
    array2 = np.array(
        ["2020-07-27T10:41:11", "2019-01-01", "2020-01-01"], "datetime64[25s]"
    )
    ak_a1 = ak.Array(array1)
    ak_a2 = ak.Array(array2)
    np_ar1 = ak_a1.to_numpy()
    np_ar2 = ak_a2.to_numpy()

    if np_ar1[0] > np_ar2[0]:
        assert (np_ar1[0] - np.timedelta64(25, "s")) < np_ar2[0]
    else:
        assert (np_ar1[0] + np.timedelta64(25, "s")) >= np_ar2[0]


def test_sum():

    dtypes = ["datetime64[s]", "timedelta64[D]"]

    arrays = (np.arange(0, 12, dtype=dtype) for dtype in dtypes)
    for array in arrays:
        content = ak.layout.NumpyArray(array)
        offsets = ak.layout.Index64(np.array([0, 4, 8, 12], dtype=np.int64))
        depth = ak.layout.ListOffsetArray64(offsets, content)

        if np.issubdtype(array.dtype, np.timedelta64):
            assert ak.to_list(depth.sum(-1)) == [
                datetime.timedelta(6),
                datetime.timedelta(22),
                datetime.timedelta(38),
            ]

            assert ak.to_list(depth.sum(1)) == [
                datetime.timedelta(6),
                datetime.timedelta(22),
                datetime.timedelta(38),
            ]

            assert ak.to_list(depth.sum(-2)) == [
                datetime.timedelta(12),
                datetime.timedelta(15),
                datetime.timedelta(18),
                datetime.timedelta(21),
            ]
            assert ak.to_list(depth.sum(0)) == [
                datetime.timedelta(12),
                datetime.timedelta(15),
                datetime.timedelta(18),
                datetime.timedelta(21),
            ]

        else:
            with pytest.raises(ValueError):
                depth.sum(-1)


def test_more():
    nparray = np.array(
        [np.datetime64("2021-06-03T10:00"), np.datetime64("2021-06-03T11:00")]
    )
    akarray = ak.Array(nparray)

    assert (akarray[1:] - akarray[:-1]).tolist() == [np.timedelta64(60, "m")]
    assert ak.sum(akarray[1:] - akarray[:-1]) == np.timedelta64(60, "m")
    assert ak.sum(akarray[1:] - akarray[:-1], axis=0) == [np.timedelta64(60, "m")]


@pytest.mark.skipif(
    ak._util.py27 or np.__version__ <= "1.13.1",
    reason="Python 2.7 module 'numpy.core' has no attribute '_exceptions'",
)
def test_ufunc_sum():
    nparray = np.array(
        [np.datetime64("2021-06-03T10:00"), np.datetime64("2021-06-03T11:00")]
    )
    akarray = ak.Array(nparray)

    with pytest.raises(np.core._exceptions.UFuncTypeError):
        akarray[1:] + akarray[:-1]


def test_ufunc_mul():
    nparray = np.array(
        [np.datetime64("2021-06-03T10:00"), np.datetime64("2021-06-03T11:00")]
    )
    akarray = ak.Array(nparray)

    with pytest.raises(ValueError):  # FIXME?: np.core._exceptions.UFuncTypeError):
        akarray * 2

    assert ak.Array([np.timedelta64(3, "D")])[0] == np.timedelta64(3, "D")


def test_NumpyArray_layout():
    array0 = ak.layout.NumpyArray(
        ["2019-09-02T09:30:00", "2019-09-13T09:30:00", "2019-09-21T20:00:00"]
    )

    assert ak.to_list(array0) == [
        "2019-09-02T09:30:00",
        "2019-09-13T09:30:00",
        "2019-09-21T20:00:00",
    ]

    array = ak.layout.NumpyArray(
        [
            np.datetime64("2019-09-02T09:30:00"),
            np.datetime64("2019-09-13T09:30:00"),
            np.datetime64("2019-09-21T20:00:00"),
        ]
    )

    assert ak.to_list(array) == [
        np.datetime64("2019-09-02T09:30:00"),
        np.datetime64("2019-09-13T09:30:00"),
        np.datetime64("2019-09-21T20:00:00"),
    ]


pandas = pytest.importorskip("pandas")


def test_from_pandas():
    values = {"time": ["20190902093000", "20190913093000", "20190921200000"]}
    df = pandas.DataFrame(values, columns=["time"])
    df["time"] = pandas.to_datetime(df["time"], format="%Y%m%d%H%M%S")
    array = ak.layout.NumpyArray(df)
    assert ak.to_list(array) == [
        np.datetime64("2019-09-02T09:30:00"),
        np.datetime64("2019-09-13T09:30:00"),
        np.datetime64("2019-09-21T20:00:00"),
    ]
    array2 = ak.Array(df.values)
    assert ak.to_list(array2) == [
        np.datetime64("2019-09-02T09:30:00"),
        np.datetime64("2019-09-13T09:30:00"),
        np.datetime64("2019-09-21T20:00:00"),
    ]


pyarrow = pytest.importorskip("pyarrow")


def test_from_arrow():
    array = ak.from_arrow(
        pyarrow.array(
            [datetime.datetime(2002, 1, 23), datetime.datetime(2019, 2, 20)],
            type=pyarrow.date64(),
        )
    )
    assert array.tolist() == [
        np.datetime64("2002-01-23T00:00:00.000"),
        np.datetime64("2019-02-20T00:00:00.000"),
    ]

    array = ak.from_arrow(
        pyarrow.array(
            [datetime.datetime(2002, 1, 23), datetime.datetime(2019, 2, 20)],
            type=pyarrow.date32(),
        )
    )
    assert array.tolist() == [
        np.datetime64("2002-01-23T00:00:00.000"),
        np.datetime64("2019-02-20T00:00:00.000"),
    ]

    array = ak.from_arrow(
        pyarrow.array(
            [datetime.time(1, 0, 0), datetime.time(2, 30, 0)],
            type=pyarrow.time64("us"),
        )
    )
    assert array.tolist() == [
        np.datetime64("1970-01-01T01:00:00.000"),
        np.datetime64("1970-01-01T02:30:00.000"),
    ]

    array = ak.from_arrow(
        pyarrow.array(
            [datetime.time(1, 0, 0), datetime.time(2, 30, 0)],
            type=pyarrow.time64("ns"),
        )
    )
    assert array.tolist() == [
        np.datetime64("1970-01-01T01:00:00.000"),
        np.datetime64("1970-01-01T02:30:00.000"),
    ]

    array = ak.from_arrow(
        pyarrow.array(
            [datetime.time(1, 0, 0), datetime.time(2, 30, 0)],
            type=pyarrow.time32("s"),
        )
    )
    assert array.tolist() == [
        np.datetime64("1970-01-01T01:00:00.000"),
        np.datetime64("1970-01-01T02:30:00.000"),
    ]

    array = ak.from_arrow(
        pyarrow.array(
            [datetime.time(1, 0, 0), datetime.time(2, 30, 0)],
            type=pyarrow.time32("ms"),
        )
    )
    assert array.tolist() == [
        np.datetime64("1970-01-01T01:00:00.000"),
        np.datetime64("1970-01-01T02:30:00.000"),
    ]

    array = ak.from_arrow(
        pyarrow.array(
            [datetime.datetime(2002, 1, 23), datetime.datetime(2019, 2, 20)],
            type=pyarrow.timestamp("s"),
        )
    )
    assert array.tolist() == [
        np.datetime64("2002-01-23T00:00:00.000"),
        np.datetime64("2019-02-20T00:00:00.000"),
    ]

    array = ak.from_arrow(
        pyarrow.array(
            [datetime.datetime(2002, 1, 23), datetime.datetime(2019, 2, 20)],
            type=pyarrow.timestamp("ms"),
        )
    )
    assert array.tolist() == [
        np.datetime64("2002-01-23T00:00:00.000"),
        np.datetime64("2019-02-20T00:00:00.000"),
    ]

    array = ak.from_arrow(
        pyarrow.array(
            [datetime.datetime(2002, 1, 23), datetime.datetime(2019, 2, 20)],
            type=pyarrow.timestamp("us"),
        )
    )
    assert array.tolist() == [
        np.datetime64("2002-01-23T00:00:00.000"),
        np.datetime64("2019-02-20T00:00:00.000"),
    ]

    array = ak.from_arrow(
        pyarrow.array(
            [datetime.datetime(2002, 1, 23), datetime.datetime(2019, 2, 20)],
            type=pyarrow.timestamp("ns"),
        )
    )
    assert array.tolist() == [
        np.datetime64("2002-01-23T00:00:00.000"),
        np.datetime64("2019-02-20T00:00:00.000"),
    ]

    array = ak.from_arrow(
        pyarrow.array(
            [datetime.timedelta(5), datetime.timedelta(10)],
            type=pyarrow.duration("s"),
        )
    )
    assert array.tolist() == [
        np.timedelta64(5, "D"),
        np.timedelta64(10, "D"),
    ]

    array = ak.from_arrow(
        pyarrow.array(
            [datetime.timedelta(5), datetime.timedelta(10)],
            type=pyarrow.duration("ms"),
        )
    )
    assert array.tolist() == [
        np.timedelta64(5, "D"),
        np.timedelta64(10, "D"),
    ]

    array = ak.from_arrow(
        pyarrow.array(
            [datetime.timedelta(5), datetime.timedelta(10)],
            type=pyarrow.duration("us"),
        )
    )
    assert array.tolist() == [
        np.timedelta64(5, "D"),
        np.timedelta64(10, "D"),
    ]

    array = ak.from_arrow(
        pyarrow.array(
            [datetime.timedelta(5), datetime.timedelta(10)],
            type=pyarrow.duration("ns"),
        )
    )
    assert array.tolist() == [
        np.timedelta64(5, "D"),
        np.timedelta64(10, "D"),
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import awkward as ak
from awkward._v2.forms.form import Form, _parameters_equal, nonvirtual
from awkward._v2.forms.indexedform import IndexedForm


class RecordForm(Form):
    def __init__(
        self,
        contents,
        keys,
        has_identifier=False,
        parameters=None,
        form_key=None,
    ):
        if not isinstance(contents, Iterable):
            raise TypeError(
                "{0} 'contents' must be iterable, not {1}".format(
                    type(self).__name__, repr(contents)
                )
            )
        for content in contents:
            if not isinstance(content, Form):
                raise TypeError(
                    "{0} all 'contents' must be Form subclasses, not {1}".format(
                        type(self).__name__, repr(content)
                    )
                )
        if keys is not None and not isinstance(keys, Iterable):
            raise TypeError(
                "{0} 'keys' must be iterable, not {1}".format(
                    type(self).__name__, repr(contents)
                )
            )

        self._keys = keys
        self._contents = list(contents)
        self._init(has_identifier, parameters, form_key)

    @property
    def keys(self):
        if self._keys is None:
            return [str(i) for i in range(len(self._contents))]
        else:
            return self._keys

    @property
    def is_tuple(self):
        return self._keys is None

    @property
    def contents(self):
        return self._contents

    def __repr__(self):
        args = [repr(self._contents), repr(self._keys)] + self._repr_args()
        return "{0}({1})".format(type(self).__name__, ", ".join(args))

    def index_to_key(self, index):
        if 0 <= index < len(self._contents):
            if self._keys is None:
                return str(index)
            else:
                return self._keys[index]
        else:
            raise IndexError(
                "no index {0} in record with {1} fields".format(
                    index, len(self._contents)
                )
            )

    def key_to_index(self, key):
        if self._keys is None:
            try:
                i = int(key)
            except ValueError:
                pass
            else:
                if 0 <= i < len(self._contents):
                    return i
        else:
            try:
                i = self._keys.index(key)
            except ValueError:
                pass
            else:
                return i
        raise IndexError(
            "no field {0} in record with {1} fields".format(
                repr(key), len(self._contents)
            )
        )

    def haskey(self, key):
        if self._keys is None:
            try:
                i = int(key)
            except ValueError:
                return False
            else:
                return 0 <= i < len(self._contents)
        else:
            return key in self._keys

    def content(self, index_or_key):
        if ak._util.isint(index_or_key):
            index = index_or_key
        elif ak._util.isstr(index_or_key):
            index = self.key_to_index(index_or_key)
        else:
            raise TypeError(
                "index_or_key must be an integer (index) or string (key), not {0}".format(
                    repr(index_or_key)
                )
            )
        return self._contents[index]

    def _tolist_part(self, verbose, toplevel):
        out = {"class": "RecordArray"}

        contents_tolist = [
            content._tolist_part(verbose, toplevel=False) for content in self._contents
        ]
        if self._keys is not None:
            out["contents"] = dict(zip(self._keys, contents_tolist))
        else:
            out["contents"] = contents_tolist

        return self._tolist_extra(out, verbose)

    def _type(self, typestrs):
        return ak._v2.types.recordtype.RecordType(
            [x._type(typestrs) for x in self._contents],
            self._keys,
            self._parameters,
            ak._util.gettypestr(self._parameters, typestrs),
        )

    def __eq__(self, other):
        if isinstance(other, RecordForm):
            if (
                self._has_identifier == other._has_identifier
                and self._form_key == other._form_key
                and self.is_tuple == other.is_tuple
                and len(self._contents) == len(other._contents)
                and _parameters_equal(self._parameters, other._parameters)
            ):
                if self.is_tuple:
                    for i in range(len(self._contents)):
                        if self._contents[i] != other._contents[i]:
                            return False
                    else:
                        return True
                else:
                    if set(self._keys) != set(other._keys):
                        return False
                    else:
                        for key, content in zip(self._keys, self._contents):
                            if content != other.content(key):
                                return False
                        else:
                            return True
            else:
                return False
        else:
            return False

    def generated_compatibility(self, other):
        other = nonvirtual(other)

        if other is None:
            return True

        elif isinstance(other, RecordForm):
            if self.is_tuple == other.is_tuple:
                self_keys = set(self.keys)
                other_keys = set(other.keys)
                if self_keys == other_keys:
                    return _parameters_equal(
                        self._parameters, other._parameters
                    ) and all(
                        self.content(x).generated_compatibility(other.content(x))
                        for x in self_keys
                    )
                else:
                    return False
            else:
                return False

        else:
            return False

    def _getitem_range(self):
        return RecordForm(
            self._contents,
            self._keys,
            has_identifier=self._has_identifier,
            parameters=self._parameters,
            form_key=None,
        )

    def _getitem_field(self, where, only_fields=()):
        if len(only_fields) == 0:
            return self.content(where)

        else:
            nexthead, nexttail = ak._v2._slicing.headtail(only_fields)
            if ak._util.isstr(nexthead):
                return self.content(where)._getitem_field(nexthead, nexttail)
            else:
                return self.content(where)._getitem_fields(nexthead, nexttail)

    def _getitem_fields(self, where, only_fields=()):
        indexes = [self.key_to_index(key) for key in where]
        if self._keys is None:
            keys = None
        else:
            keys = [self._keys[i] for i in indexes]

        if len(only_fields) == 0:
            contents = [self.content(i) for i in indexes]
        else:
            nexthead, nexttail = ak._v2._slicing.headtail(only_fields)
            if ak._util.isstr(nexthead):
                contents = [
                    self.content(i)._getitem_field(nexthead, nexttail) for i in indexes
                ]
            else:
                contents = [
                    self.content(i)._getitem_fields(nexthead, nexttail) for i in indexes
                ]

        return RecordForm(
            contents,
            keys,
            has_identifier=self._has_identifier,
            parameters=None,
            form_key=None,
        )

    def _carry(self, allow_lazy):
        if allow_lazy:
            return IndexedForm(
                "i64",
                self,
                has_identifier=self._has_identifier,
                parameters=None,
                form_key=None,
            )
        else:
            return RecordForm(
                self._contents,
                self._keys,
                has_identifier=self._has_identifier,
                parameters=self._parameters,
                form_key=None,
            )

    @property
    def purelist_isregular(self):
        return True

    @property
    def purelist_depth(self):
        return 1

    @property
    def minmax_depth(self):
        if len(self._contents) == 0:
            return (0, 0)
        mins, maxs = [], []
        for content in self._contents:
            mindepth, maxdepth = content.minmax_depth
            mins.append(mindepth)
            maxs.append(maxdepth)
        return (min(mins), max(maxs))

    @property
    def branch_depth(self):
        if len(self._contents) == 0:
            return (False, 1)
        anybranch = False
        mindepth = None
        for content in self._contents:
            branch, depth = content.branch_depth
            if mindepth is None:
                mindepth = depth
            if branch or mindepth != depth:
                anybranch = True
            if mindepth > depth:
                mindepth = depth
        return (anybranch, mindepth)

    @property
    def dimension_optiontype(self):
        return False



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    assert isinstance(ak.zip({"x": 1, "y": 0, "z": 0}), ak.Record)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward._v2.tmp_for_testing import v1_to_v2

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)

content = ak.layout.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
starts1 = ak.layout.IndexU32(np.array([0, 3, 3, 5, 6], np.uint32))
stops1 = ak.layout.IndexU32(np.array([3, 3, 5, 6, 9], np.uint32))
offsets1 = ak.layout.IndexU32(np.array([0, 3, 3, 5, 6, 9], np.uint32))
starts2 = ak.layout.IndexU32(np.array([0, 2, 3, 3], np.uint32))
stops2 = ak.layout.IndexU32(np.array([2, 3, 3, 5], np.uint32))
offsets2 = ak.layout.IndexU32(np.array([0, 2, 3, 3, 5], np.uint32))


def test_listarray_basic():
    array1 = ak.layout.ListArrayU32(starts1, stops1, content)
    array2 = ak.layout.ListArrayU32(starts2, stops2, array1)
    array1 = v1_to_v2(array1)
    array2 = v1_to_v2(array2)

    assert ak.to_list(array1) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]
    assert ak.to_list(array1[2]) == [4.4, 5.5]
    assert array1.typetracer[2].form == array1[2].form
    assert ak.to_list(array1[1:-1]) == [[], [4.4, 5.5], [6.6]]
    assert array1.typetracer[1:-1].form == array1[1:-1].form
    assert ak.to_list(array2) == [
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [],
        [[6.6], [7.7, 8.8, 9.9]],
    ]
    assert ak.to_list(array2[1]) == [[4.4, 5.5]]
    assert array2.typetracer[1].form == array2[1].form
    assert ak.to_list(array2[1:-1]) == [[[4.4, 5.5]], []]
    assert array2.typetracer[1:-1].form == array2[1:-1].form


def test_listoffsetarray_basic():
    array1 = ak.layout.ListOffsetArrayU32(offsets1, content)
    array2 = ak.layout.ListOffsetArrayU32(offsets2, array1)
    array1 = v1_to_v2(array1)
    array2 = v1_to_v2(array2)

    assert ak.to_list(array1) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9],
    ]
    assert ak.to_list(array1[2]) == [4.4, 5.5]
    assert array1.typetracer[2].form == array1[2].form
    assert ak.to_list(array1[1:-1]) == [[], [4.4, 5.5], [6.6]]
    assert array1.typetracer[1:-1].form == array1[1:-1].form
    assert ak.to_list(array2) == [
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [],
        [[6.6], [7.7, 8.8, 9.9]],
    ]
    assert ak.to_list(array2[1]) == [[4.4, 5.5]]
    assert array2.typetracer[1].form == array2[1].form
    assert ak.to_list(array2[1:-1]) == [[[4.4, 5.5]], []]
    assert array2.typetracer[1:-1].form == array2[1:-1].form


def test_listarray_at():
    array1 = ak.layout.ListArrayU32(starts1, stops1, content)
    array1 = v1_to_v2(array1)

    assert ak.to_list(array1[2]) == [4.4, 5.5]
    assert array1.typetracer[2].form == array1[2].form
    assert (
        ak.to_list(
            array1[
                2,
            ]
        )
        == [4.4, 5.5]
    )
    assert (
        array1.typetracer[
            2,
        ].form
        == array1[
            2,
        ].form
    )
    assert ak.to_list(array1[2, 1:]) == [5.5]
    assert array1.typetracer[2, 1:].form == array1[2, 1:].form
    assert ak.to_list(array1[2:, 0]) == [4.4, 6.6, 7.7]
    assert array1.typetracer[2:, 0].form == array1[2:, 0].form
    assert ak.to_list(array1[2:, -1]) == [5.5, 6.6, 9.9]
    assert array1.typetracer[2:, -1].form == array1[2:, -1].form


def test_listoffsetarray_at():
    array1 = ak.layout.ListOffsetArrayU32(offsets1, content)
    array1 = v1_to_v2(array1)
    assert (
        ak.to_list(
            array1[
                2,
            ]
        )
        == [4.4, 5.5]
    )
    assert (
        array1.typetracer[
            2,
        ].form
        == array1[
            2,
        ].form
    )
    assert ak.to_list(array1[2, 1:]) == [5.5]
    assert array1.typetracer[2, 1:].form == array1[2, 1:].form
    assert ak.to_list(array1[2:, 0]) == [4.4, 6.6, 7.7]
    assert array1.typetracer[2:, 0].form == array1[2:, 0].form
    assert ak.to_list(array1[2:, -1]) == [5.5, 6.6, 9.9]
    assert array1.typetracer[2:, -1].form == array1[2:, -1].form


def test_listarray_slice():
    array1 = ak.layout.ListArrayU32(starts1, stops1, content)
    array2 = ak.layout.ListArrayU32(starts2, stops2, array1)
    array1 = v1_to_v2(array1)
    array2 = v1_to_v2(array2)

    assert ak.to_list(array1[1:-1]) == [[], [4.4, 5.5], [6.6]]
    assert array1.typetracer[1:-1].form == array1[1:-1].form
    assert (
        ak.to_list(
            array1[
                1:-1,
            ]
        )
        == [[], [4.4, 5.5], [6.6]]
    )
    assert (
        array1.typetracer[
            1:-1,
        ].form
        == array1[
            1:-1,
        ].form
    )
    assert ak.to_list(array2[1:-1]) == [[[4.4, 5.5]], []]
    assert array2.typetracer[1:-1].form == array2[1:-1].form
    assert (
        ak.to_list(
            array2[
                1:-1,
            ]
        )
        == [[[4.4, 5.5]], []]
    )
    assert (
        array2.typetracer[
            1:-1,
        ].form
        == array2[
            1:-1,
        ].form
    )


def test_listoffsetarray_slice():
    array1 = ak.layout.ListOffsetArrayU32(offsets1, content)
    array2 = ak.layout.ListOffsetArrayU32(offsets2, array1)
    array1 = v1_to_v2(array1)
    array2 = v1_to_v2(array2)

    assert ak.to_list(array1[1:-1]) == [[], [4.4, 5.5], [6.6]]
    assert array1.typetracer[1:-1].form == array1[1:-1].form
    assert (
        ak.to_list(
            array1[
                1:-1,
            ]
        )
        == [[], [4.4, 5.5], [6.6]]
    )
    assert (
        array1.typetracer[
            1:-1,
        ].form
        == array1[
            1:-1,
        ].form
    )
    assert ak.to_list(array2[1:-1]) == [[[4.4, 5.5]], []]
    assert array2.typetracer[1:-1].form == array2[1:-1].form
    assert (
        ak.to_list(
            array2[
                1:-1,
            ]
        )
        == [[[4.4, 5.5]], []]
    )
    assert (
        array2.typetracer[
            1:-1,
        ].form
        == array2[
            1:-1,
        ].form
    )


def test_listarray_slice_slice():
    array1 = ak.layout.ListArrayU32(starts1, stops1, content)
    array1 = v1_to_v2(array1)

    assert ak.to_list(array1[2:]) == [[4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert array1.typetracer[2:].form == array1[2:].form
    assert ak.to_list(array1[2:, 1:]) == [[5.5], [], [8.8, 9.9]]
    assert array1.typetracer[2:, 1:].form == array1[2:, 1:].form
    assert ak.to_list(array1[2:, :-1]) == [[4.4], [], [7.7, 8.8]]
    assert array1.typetracer[2:, :-1].form == array1[2:, :-1].form


def test_listoffsetarray_slice_slice():
    array1 = ak.layout.ListOffsetArrayU32(offsets1, content)
    array1 = v1_to_v2(array1)

    assert ak.to_list(array1[2:]) == [[4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert array1.typetracer[2:].form == array1[2:].form
    assert ak.to_list(array1[2:, 1:]) == [[5.5], [], [8.8, 9.9]]
    assert array1.typetracer[2:, 1:].form == array1[2:, 1:].form
    assert ak.to_list(array1[2:, :-1]) == [[4.4], [], [7.7, 8.8]]
    assert array1.typetracer[2:, :-1].form == array1[2:, :-1].form


def test_listarray_ellipsis():
    array1 = ak.layout.ListArrayU32(starts1, stops1, content)
    array2 = ak.layout.ListArrayU32(starts2, stops2, array1)
    array1 = v1_to_v2(array1)
    array2 = v1_to_v2(array2)

    if not ak._util.py27:
        assert ak.to_list(array1[Ellipsis, 1:]) == [
            [2.2, 3.3],
            [],
            [5.5],
            [],
            [8.8, 9.9],
        ]
        assert array1.typetracer[Ellipsis, 1:].form == array1[Ellipsis, 1:].form
        assert ak.to_list(array2[Ellipsis, 1:]) == [
            [[2.2, 3.3], []],
            [[5.5]],
            [],
            [[], [8.8, 9.9]],
        ]
        assert array2.typetracer[Ellipsis, 1:].form == array2[Ellipsis, 1:].form


def test_listoffsetarray_ellipsis():
    array1 = ak.layout.ListOffsetArrayU32(offsets1, content)
    array2 = ak.layout.ListOffsetArrayU32(offsets2, array1)
    array1 = v1_to_v2(array1)
    array2 = v1_to_v2(array2)

    if not ak._util.py27:
        assert ak.to_list(array1[Ellipsis, 1:]) == [
            [2.2, 3.3],
            [],
            [5.5],
            [],
            [8.8, 9.9],
        ]
        assert array1.typetracer[Ellipsis, 1:].form == array1[Ellipsis, 1:].form
        assert ak.to_list(array2[Ellipsis, 1:]) == [
            [[2.2, 3.3], []],
            [[5.5]],
            [],
            [[], [8.8, 9.9]],
        ]
        assert array2.typetracer[Ellipsis, 1:].form == array2[Ellipsis, 1:].form


def test_listarray_array_slice():
    array1 = ak.layout.ListArrayU32(starts1, stops1, content)
    array2 = ak.layout.ListArrayU32(starts2, stops2, array1)
    array1 = v1_to_v2(array1)
    array2 = v1_to_v2(array2)

    assert ak.to_list(array2[[0, 0, 1, 1, 1, 0]]) == [
        [[1.1, 2.2, 3.3], []],
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[1.1, 2.2, 3.3], []],
    ]
    assert array2.typetracer[[0, 0, 1, 1, 1, 0]].form == array2[[0, 0, 1, 1, 1, 0]].form
    assert ak.to_list(array2[[0, 0, 1, 1, 1, 0], :]) == [
        [[1.1, 2.2, 3.3], []],
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[1.1, 2.2, 3.3], []],
    ]
    assert (
        array2.typetracer[[0, 0, 1, 1, 1, 0], :].form
        == array2[[0, 0, 1, 1, 1, 0], :].form
    )
    assert ak.to_list(array2[[0, 0, 1, 1, 1, 0], :, 1:]) == [
        [[2.2, 3.3], []],
        [[2.2, 3.3], []],
        [[5.5]],
        [[5.5]],
        [[5.5]],
        [[2.2, 3.3], []],
    ]
    assert (
        array2.typetracer[[0, 0, 1, 1, 1, 0], :, 1:].form
        == array2[[0, 0, 1, 1, 1, 0], :, 1:].form
    )


def test_listoffsetarray_array_slice():
    array1 = ak.layout.ListOffsetArrayU32(offsets1, content)
    array2 = ak.layout.ListOffsetArrayU32(offsets2, array1)
    array1 = v1_to_v2(array1)
    array2 = v1_to_v2(array2)

    assert ak.to_list(array2[[0, 0, 1, 1, 1, 0]]) == [
        [[1.1, 2.2, 3.3], []],
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[1.1, 2.2, 3.3], []],
    ]
    assert array2.typetracer[[0, 0, 1, 1, 1, 0]].form == array2[[0, 0, 1, 1, 1, 0]].form
    assert ak.to_list(array2[[0, 0, 1, 1, 1, 0], :]) == [
        [[1.1, 2.2, 3.3], []],
        [[1.1, 2.2, 3.3], []],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[4.4, 5.5]],
        [[1.1, 2.2, 3.3], []],
    ]
    assert (
        array2.typetracer[[0, 0, 1, 1, 1, 0], :].form
        == array2[[0, 0, 1, 1, 1, 0], :].form
    )
    assert ak.to_list(array2[[0, 0, 1, 1, 1, 0], :, 1:]) == [
        [[2.2, 3.3], []],
        [[2.2, 3.3], []],
        [[5.5]],
        [[5.5]],
        [[5.5]],
        [[2.2, 3.3], []],
    ]
    assert (
        array2.typetracer[[0, 0, 1, 1, 1, 0], :, 1:].form
        == array2[[0, 0, 1, 1, 1, 0], :, 1:].form
    )


def test_listarray_array():
    array1 = ak.layout.ListArrayU32(starts1, stops1, content)
    array1 = v1_to_v2(array1)

    assert ak.to_list(array1[np.array([2, 0, 0, 1, -1])]) == [
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        [1.1, 2.2, 3.3],
        [],
        [7.7, 8.8, 9.9],
    ]
    assert (
        array1.typetracer[np.array([2, 0, 0, 1, -1])].form
        == array1[np.array([2, 0, 0, 1, -1])].form
    )
    assert ak.to_list(array1[np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0])]) == [
        5.5,
        2.2,
        1.1,
        7.7,
    ]
    assert (
        array1.typetracer[np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0])].form
        == array1[np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0])].form
    )

    content_deep = ak.layout.NumpyArray(
        np.array(
            [
                [0, 0],
                [1, 10],
                [2, 20],
                [3, 30],
                [4, 40],
                [5, 50],
                [6, 60],
                [7, 70],
                [8, 80],
            ]
        )
    )
    starts1_deep = ak.layout.IndexU32(np.array([0, 3, 6]))
    stops1_deep = ak.layout.IndexU32(np.array([3, 6, 9]))
    array1_deep = ak.layout.ListArrayU32(starts1_deep, stops1_deep, content_deep)
    array1_deep = v1_to_v2(array1_deep)

    assert ak.to_list(array1_deep) == [
        [[0, 0], [1, 10], [2, 20]],
        [[3, 30], [4, 40], [5, 50]],
        [[6, 60], [7, 70], [8, 80]],
    ]
    s = (np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0]), np.array([0, 1, 0, 1]))
    assert (
        np.array(
            [
                [[0, 0], [1, 10], [2, 20]],
                [[3, 30], [4, 40], [5, 50]],
                [[6, 60], [7, 70], [8, 80]],
            ]
        )[s].tolist()
        == ak.to_list(array1_deep[s])
    )
    assert array1_deep.typetracer[s].form == array1_deep[s].form

    s = (np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0]), slice(1, None))
    assert (
        np.array(
            [
                [[0, 0], [1, 10], [2, 20]],
                [[3, 30], [4, 40], [5, 50]],
                [[6, 60], [7, 70], [8, 80]],
            ]
        )[s].tolist()
        == ak.to_list(array1_deep[s])
    )
    assert array1_deep.typetracer[s].form == array1_deep[s].form


def test_listoffsetarray_array():
    array1 = ak.layout.ListOffsetArrayU32(offsets1, content)
    array1 = v1_to_v2(array1)

    assert ak.to_list(array1[np.array([2, 0, 0, 1, -1])]) == [
        [4.4, 5.5],
        [1.1, 2.2, 3.3],
        [1.1, 2.2, 3.3],
        [],
        [7.7, 8.8, 9.9],
    ]
    assert (
        array1.typetracer[np.array([2, 0, 0, 1, -1])].form
        == array1[np.array([2, 0, 0, 1, -1])].form
    )
    assert ak.to_list(array1[np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0])]) == [
        5.5,
        2.2,
        1.1,
        7.7,
    ]
    assert (
        array1.typetracer[np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0])].form
        == array1[np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0])].form
    )

    content_deep = ak.layout.NumpyArray(
        np.array(
            [
                [0, 0],
                [1, 10],
                [2, 20],
                [3, 30],
                [4, 40],
                [5, 50],
                [6, 60],
                [7, 70],
                [8, 80],
            ]
        )
    )
    starts1_deep = ak.layout.IndexU32(np.array([0, 3, 6]))
    stops1_deep = ak.layout.IndexU32(np.array([3, 6, 9]))
    array1_deep = ak.layout.ListArrayU32(starts1_deep, stops1_deep, content_deep)
    array1_deep = v1_to_v2(array1_deep)

    assert ak.to_list(array1_deep) == [
        [[0, 0], [1, 10], [2, 20]],
        [[3, 30], [4, 40], [5, 50]],
        [[6, 60], [7, 70], [8, 80]],
    ]
    s = (np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0]), np.array([0, 1, 0, 1]))
    assert (
        np.array(
            [
                [[0, 0], [1, 10], [2, 20]],
                [[3, 30], [4, 40], [5, 50]],
                [[6, 60], [7, 70], [8, 80]],
            ]
        )[s].tolist()
        == ak.to_list(array1_deep[s])
    )
    assert array1_deep.typetracer[s].form == array1_deep[s].form

    s = (np.array([2, 0, 0, -1]), np.array([1, 1, 0, 0]), slice(1, None))
    assert (
        np.array(
            [
                [[0, 0], [1, 10], [2, 20]],
                [[3, 30], [4, 40], [5, 50]],
                [[6, 60], [7, 70], [8, 80]],
            ]
        )[s].tolist()
        == ak.to_list(array1_deep[s])
    )
    assert array1_deep.typetracer[s].form == array1_deep[s].form



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

from awkward._ext import fromjson
from awkward._ext import uproot_issue_90


__all__ = ["fromjson", "uproot_issue_90"]


def __dir__():
    return __all__



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


pyarrow = pytest.importorskip("pyarrow")


def test_categorical_is_valid():
    # validate a categorical array by its content
    arr = ak.Array([2019, 2020, 2021, 2020, 2019])
    categorical = ak.to_categorical(arr)
    assert ak.is_valid(categorical)


def test_optional_categorical_from_arrow():
    # construct categorical array from option-typed DictionaryArray
    indices = pyarrow.array([0, 1, 0, 1, 2, 0, 2])
    nan_indices = pyarrow.array([0, 1, 0, 1, 2, None, 0, 2])
    dictionary = pyarrow.array([2019, 2020, 2021])

    dict_array = pyarrow.DictionaryArray.from_arrays(indices, dictionary)
    categorical_array = ak.from_arrow(dict_array)
    assert categorical_array.layout.parameter("__array__") == "categorical"

    option_dict_array = pyarrow.DictionaryArray.from_arrays(nan_indices, dictionary)
    option_categorical_array = ak.from_arrow(option_dict_array)
    assert option_categorical_array.layout.parameter("__array__") == "categorical"


def test_categorical_from_arrow_ChunkedArray():
    indices = [0, 1, 0, 1, 2, 0, 2]
    indices_new_schema = [0, 1, 0, 1, 0]

    dictionary = pyarrow.array([2019, 2020, 2021])
    dictionary_new_schema = pyarrow.array([2019, 2020])

    dict_array = pyarrow.DictionaryArray.from_arrays(pyarrow.array(indices), dictionary)
    dict_array_new_schema = pyarrow.DictionaryArray.from_arrays(
        pyarrow.array(indices_new_schema), dictionary_new_schema
    )

    batch = pyarrow.RecordBatch.from_arrays([dict_array], ["year"])
    batch_new_schema = pyarrow.RecordBatch.from_arrays(
        [dict_array_new_schema], ["year"]
    )

    batches = [batch] * 3
    batches_mixed_schema = [batch] + [batch_new_schema]

    table = pyarrow.Table.from_batches(batches)
    table_mixed_schema = pyarrow.Table.from_batches(batches_mixed_schema)

    array = ak.from_arrow(table)
    array_mixed_schema = ak.from_arrow(table_mixed_schema)

    assert np.asarray(array.layout.field(0).content.index).tolist() == indices * 3
    assert (
        np.asarray(array_mixed_schema.layout.field(0).content.index).tolist()
        == indices + indices_new_schema
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward._v2.tmp_for_testing import v1_to_v2

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)


def test_numpyarray_getitem_bystrides():
    a = np.arange(10)
    b = ak.layout.NumpyArray(a)
    b = v1_to_v2(b)

    assert b[3] == a[3]
    assert b[-3] == a[-3]
    assert ak.to_list(b[()]) == ak.to_list(a[()])
    assert ak.to_list(b[slice(None)]) == ak.to_list(a[slice(None)])
    assert ak.to_list(b[slice(3, 7)]) == ak.to_list(a[slice(3, 7)])
    assert ak.to_list(b[slice(3, 100)]) == ak.to_list(a[slice(3, 100)])
    assert ak.to_list(b[slice(-100, 7)]) == ak.to_list(a[slice(-100, 7)])
    assert ak.to_list(b[slice(3, -3)]) == ak.to_list(a[slice(3, -3)])
    assert ak.to_list(b[slice(1, 7, 2)]) == ak.to_list(a[slice(1, 7, 2)])
    assert ak.to_list(b[slice(-8, 7, 2)]) == ak.to_list(a[slice(-8, 7, 2)])
    assert ak.to_list(b[slice(None, 7, 2)]) == ak.to_list(a[slice(None, 7, 2)])
    assert ak.to_list(b[slice(None, -3, 2)]) == ak.to_list(a[slice(None, -3, 2)])
    assert ak.to_list(b[slice(8, None, -1)]) == ak.to_list(a[slice(8, None, -1)])
    assert ak.to_list(b[slice(8, None, -2)]) == ak.to_list(a[slice(8, None, -2)])
    assert ak.to_list(b[slice(-2, None, -2)]) == ak.to_list(a[slice(-2, None, -2)])

    a = np.arange(7 * 5).reshape(7, 5)
    b = ak.layout.NumpyArray(a)
    b = v1_to_v2(b)

    assert ak.to_list(b[()]) == ak.to_list(a[()])
    assert ak.to_list(b[3]) == ak.to_list(a[3])
    assert ak.to_list(b[(3, 2)]) == ak.to_list(a[3, 2])
    assert ak.to_list(b[slice(1, 4)]) == ak.to_list(a[slice(1, 4)])
    assert ak.to_list(b[(3, slice(1, 4))]) == ak.to_list(a[3, slice(1, 4)])
    assert ak.to_list(b[(slice(1, 4), 3)]) == ak.to_list(a[slice(1, 4), 3])
    assert ak.to_list(b[(slice(1, 4), slice(2, None))]) == ak.to_list(
        a[slice(1, 4), slice(2, None)]
    )
    assert ak.to_list(b[(slice(None, None, -1), slice(2, None))]) == ak.to_list(
        a[slice(None, None, -1), slice(2, None)]
    )

    assert ak.to_list(b[:, np.newaxis, :]) == ak.to_list(a[:, np.newaxis, :])
    assert ak.to_list(b[np.newaxis, :, np.newaxis, :, np.newaxis]) == ak.to_list(
        a[np.newaxis, :, np.newaxis, :, np.newaxis]
    )

    if not ak._util.py27:
        assert ak.to_list(b[Ellipsis, 3]) == ak.to_list(a[Ellipsis, 3])
        assert ak.to_list(b[Ellipsis, 3, 2]) == ak.to_list(a[Ellipsis, 3, 2])
        assert ak.to_list(b[3, Ellipsis]) == ak.to_list(a[3, Ellipsis])
        assert ak.to_list(b[3, 2, Ellipsis]) == ak.to_list(a[3, 2, Ellipsis])


def test_numpyarray_getitem_next():
    a = np.arange(10)
    b = ak.layout.NumpyArray(a)
    c = np.array([7, 3, 3, 5])
    b = v1_to_v2(b)

    assert ak.to_list(b[c]) == ak.to_list(a[c])
    assert b.typetracer[c].form == b[c].form
    c = np.array([0, 0, 0, 1, 1, 1, 0, 1, 0, 1])
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    assert b.typetracer[c].form == b[c].form
    c = np.array([False, False, False, True, True, True, False, True, False, True])
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    assert b.typetracer[c].form == b[c].form
    c = np.array([], dtype=int)
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    assert b.typetracer[c].form == b[c].form

    a = np.arange(10 * 3).reshape(10, 3)
    b = ak.layout.NumpyArray(a)
    c = np.array([7, 3, 3, 5])
    b = v1_to_v2(b)

    assert ak.to_list(b[c]) == ak.to_list(a[c])
    assert b.typetracer[c].form == b[c].form
    c = np.array([False, False, False, True, True, True, False, True, False, True])
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    assert b.typetracer[c].form == b[c].form
    c = np.array([], dtype=int)
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    assert b.typetracer[c].form == b[c].form

    a = np.arange(7 * 5).reshape(7, 5)
    b = ak.layout.NumpyArray(a)
    b = v1_to_v2(b)

    c1 = np.array([], np.int64)
    c2 = np.array([], np.int64)

    assert ak.to_list(b[c1, c2]) == ak.to_list(a[c1, c2])
    assert ak.Array(b[c1, c2]).ndim == a[c1, c2].ndim
    assert b.typetracer[c1, c2].form == b[c1, c2].form

    a = np.arange(7 * 5).reshape(7, 5)
    b = ak.layout.NumpyArray(a)
    b = v1_to_v2(b)
    c1 = np.array([4, 1, 1, 3])
    c2 = np.array([2, 2, 0, 1])
    assert ak.to_list(b[c1, c2]) == ak.to_list(a[c1, c2])
    assert b.typetracer[c1, c2].form == b[c1, c2].form

    c = np.array([False, False, True, True, False, True, True])
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    assert b.typetracer[c].form == b[c].form

    c = np.array([], dtype=int)
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    c1 = np.array([], dtype=int)
    c2 = np.array([], dtype=int)
    assert ak.to_list(b[c1, c2]) == ak.to_list(a[c1, c2])
    assert b.typetracer[c1, c2].form == b[c1, c2].form

    a = np.arange(7 * 5).reshape(7, 5)
    b = ak.layout.NumpyArray(a)
    b = v1_to_v2(b)
    c = np.array([2, 0, 0, 1])
    assert ak.to_list(b[1:4, c]) == ak.to_list(a[1:4, c])
    assert ak.to_list(b[c, 1:4]) == ak.to_list(a[c, 1:4])
    assert ak.to_list(b[1:4, np.newaxis, c]) == ak.to_list(a[1:4, np.newaxis, c])
    assert ak.to_list(b[c, np.newaxis, 1:4]) == ak.to_list(a[c, np.newaxis, 1:4])
    assert ak.to_list(b[1:4, np.newaxis, np.newaxis, c]) == ak.to_list(
        a[1:4, np.newaxis, np.newaxis, c]
    )
    assert ak.to_list(b[c, np.newaxis, np.newaxis, 1:4]) == ak.to_list(
        a[c, np.newaxis, np.newaxis, 1:4]
    )
    if not ak._util.py27:
        assert ak.to_list(b[Ellipsis, c]) == ak.to_list(a[Ellipsis, c])
        assert ak.to_list(b[c, Ellipsis]) == ak.to_list(a[c, Ellipsis])


def test_numpyarray_getitem_next_2():
    a = np.arange(7 * 5).reshape(7, 5)
    b = ak.layout.NumpyArray(a)
    b = v1_to_v2(b)

    c1 = np.array([[4, 1], [1, 3], [0, 4]])
    c2 = np.array([[2, 2], [0, 1], [1, 3]])
    assert ak.to_list(b[c1, c2]) == ak.to_list(a[c1, c2])

    c = a % 2 == 0  # two dimensional
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    assert ak.to_list(b[c,]) == ak.to_list(  # noqa: E231
        a[
            c,
        ]
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


pyarrow = pytest.importorskip("pyarrow")


def test():
    assert ak.from_arrow(
        pyarrow.Table.from_pydict({"x": [None, None, None]})
    ).tolist() == [{"x": None}, {"x": None}, {"x": None}]
    assert ak.from_arrow(
        pyarrow.Table.from_pydict({"x": [[None, None], [], [None]]})
    ).tolist() == [{"x": [None, None]}, {"x": []}, {"x": [None]}]
